import threading
import warnings
from functools import cache

from typing_extensions import deprecated

try:
    import gevent
except ImportError:
    gevent = None


__all__ = (
    "enable_gevent",  # deprecated
    "greenlet_get_ident",
    "is_gevent_enabled",  # deprecated
    "new_fine_grained_lock",  # deprecated
    "thread_get_ident",
    "thread_get_name",
)


_gevent_enabled = False


@deprecated("enable_gevent() is deprecated, use gevent.monkey instead.")
def enable_gevent():
    global _gevent_enabled
    if gevent is not None:
        _gevent_enabled = True


def _disable_gevent():  # for testing
    global _gevent_enabled
    _gevent_enabled = False


@deprecated("is_gevent_enabled() is deprecated")
def is_gevent_enabled():
    return _gevent_enabled


def thread_get_name():
    return threading.current_thread().name


if gevent is None:
    greenlet_get_ident = threading.get_ident
    thread_get_ident = threading.get_ident
else:
    from gevent.monkey import get_original

    greenlet_get_ident = threading.get_ident
    thread_get_ident = get_original("threading", "get_ident")


@cache
def _get_gevent_lock_cls():
    import gevent.lock
    from gevent.monkey import is_module_patched

    if is_module_patched("threading"):
        return threading.RLock
    else:
        warnings.warn(
            (
                "Implicit use of gevent.lock.RLock via logbook.concurrency."
                "enable_gevent() is deprecated. Use gevent.monkey.patch_all() "
                "or gevent.monkey.patch_thread() instead."
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )
        return gevent.lock.RLock


def _new_fine_grained_lock():
    global _gevent_enabled
    if _gevent_enabled:
        lock_cls = _get_gevent_lock_cls()
        return lock_cls()
    else:
        return threading.RLock()


@deprecated("new_fine_grained_lock() is deprecated")
def new_fine_grained_lock():
    return _new_fine_grained_lock()



"""
logbook.queues
~~~~~~~~~~~~~~

This module implements queue backends.

:copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
:license: BSD, see LICENSE for more details.
"""

import json
import platform
import threading
from queue import Empty, Full
from queue import Queue as ThreadQueue
from threading import Lock, Thread

from logbook.base import NOTSET, LogRecord, dispatch_record
from logbook.handlers import Handler, WrapperHandler


class RedisHandler(Handler):
    """A handler that sends log messages to a Redis instance.

    It publishes each record as json dump. Requires redis module.

    To receive such records you need to have a running instance of Redis.

    Example setup::

        handler = RedisHandler("http://127.0.0.1", port="9200", key="redis")

    If your Redis instance is password protected, you can securely connect
    passing your password when creating a RedisHandler object.

    Example::

        handler = RedisHandler(password="your_redis_password")

    More info about the default buffer size: wp.me/p3tYJu-3b
    """

    def __init__(
        self,
        host="127.0.0.1",
        port=6379,
        key="redis",
        extra_fields=None,
        flush_threshold=128,
        flush_time=1,
        level=NOTSET,
        filter=None,
        password=False,
        bubble=True,
        context=None,
        push_method="rpush",
    ):
        Handler.__init__(self, level, filter, bubble)
        try:
            import redis
            from redis import ResponseError
        except ImportError:
            raise RuntimeError("The redis library is required for the RedisHandler")

        self.redis = redis.Redis(
            host=host, port=port, password=password, decode_responses=True
        )
        try:
            self.redis.ping()
        except ResponseError:
            raise ResponseError("The password provided is apparently incorrect")
        self.key = key
        self.extra_fields = extra_fields or {}
        self.flush_threshold = flush_threshold
        self.queue = []
        self.lock = Lock()
        self.push_method = push_method

        # Set up a thread that flushes the queue every specified seconds
        self._stop_event = threading.Event()
        self._flushing_t = threading.Thread(
            target=self._flush_task, args=(flush_time, self._stop_event)
        )
        self._flushing_t.daemon = True
        self._flushing_t.start()

    def _flush_task(self, time, stop_event):
        """Calls the method _flush_buffer every certain time."""
        while not self._stop_event.is_set():
            with self.lock:
                self._flush_buffer()
            self._stop_event.wait(time)

    def _flush_buffer(self):
        """Flushes the messaging queue into Redis.

        All values are pushed at once for the same key.

        The method rpush/lpush is defined by push_method argument
        """
        if self.queue:
            getattr(self.redis, self.push_method)(self.key, *self.queue)
        self.queue = []

    def disable_buffering(self):
        """Disables buffering.

        If called, every single message will be directly pushed to Redis.
        """
        self._stop_event.set()
        self.flush_threshold = 1

    def emit(self, record):
        """Emits a pair (key, value) to redis.

        The key is the one provided when creating the handler, or redis if none
        was provided. The value contains both the message and the hostname.
        Extra values are also appended to the message.
        """
        with self.lock:
            r = {
                "message": record.msg,
                "host": platform.node(),
                "level": record.level_name,
                "time": record.time.isoformat(),
            }
            r.update(self.extra_fields)
            r.update(record.kwargs)
            self.queue.append(json.dumps(r))
            if len(self.queue) == self.flush_threshold:
                self._flush_buffer()

    def close(self):
        self._flush_buffer()


class MessageQueueHandler(Handler):
    """A handler that acts as a message queue publisher, which publishes each
    record as json dump. Requires the kombu module.

    The queue will be filled with JSON exported log records.  To receive such
    log records from a queue you can use the :class:`MessageQueueSubscriber`.

    For an AMQP backend such as RabbitMQ::

        handler = MessageQueueHandler("amqp://guest:guest@localhost//")

    This requires the py-amqp or the librabbitmq client library.

    For Redis (requires redis client library)::

        handler = MessageQueueHandler("redis://localhost:8889/0")

    For MongoDB (requires pymongo)::

        handler = MessageQueueHandler("mongodb://localhost:27017/logging")

    Several other backends are also supported.
    Refer to the `kombu`_ documentation

    .. _kombu: https://docs.celeryq.dev/projects/kombu/en/latest/introduction.html
    """

    def __init__(
        self, uri=None, queue="logging", level=NOTSET, filter=None, bubble=False
    ):
        Handler.__init__(self, level, filter, bubble)
        try:
            import kombu
        except ImportError:
            raise RuntimeError(
                "The kombu library is required for the RabbitMQSubscriber."
            )
        if uri:
            connection = kombu.Connection(uri)

        self.queue = connection.SimpleQueue(queue)

    def export_record(self, record):
        """Exports the record into a dictionary ready for JSON dumping."""
        return record.to_dict(json_safe=True)

    def emit(self, record):
        self.queue.put(self.export_record(record))

    def close(self):
        self.queue.close()


RabbitMQHandler = MessageQueueHandler


class ZeroMQHandler(Handler):
    """A handler that acts as a ZeroMQ publisher, which publishes each record
    as json dump.  Requires the pyzmq library.

    The queue will be filled with JSON exported log records.  To receive such
    log records from a queue you can use the :class:`ZeroMQSubscriber`.

    If `multi` is set to `True`, the handler will use a `PUSH` socket to
    publish the records. This allows multiple handlers to use the same `uri`.
    The records can be received by using the :class:`ZeroMQSubscriber` with
    `multi` set to `True`.


    Example setup::

        handler = ZeroMQHandler("tcp://127.0.0.1:5000")
    """

    def __init__(
        self,
        uri=None,
        level=NOTSET,
        filter=None,
        bubble=False,
        context=None,
        multi=False,
    ):
        Handler.__init__(self, level, filter, bubble)
        try:
            import zmq
        except ImportError:
            raise RuntimeError("The pyzmq library is required for the ZeroMQHandler.")
        #: the zero mq context
        self.context = context or zmq.Context()

        if multi:
            #: the zero mq socket.
            self.socket = self.context.socket(zmq.PUSH)
            if uri is not None:
                self.socket.connect(uri)
        else:
            #: the zero mq socket.
            self.socket = self.context.socket(zmq.PUB)
            if uri is not None:
                self.socket.bind(uri)

    def export_record(self, record):
        """Exports the record into a dictionary ready for JSON dumping."""
        return record.to_dict(json_safe=True)

    def emit(self, record):
        self.socket.send(json.dumps(self.export_record(record)).encode("utf-8"))

    def close(self, linger=-1):
        self.socket.close(linger)

    def __del__(self):
        # When the Handler is deleted we must close our socket in a
        # non-blocking fashion (using linger).
        # Otherwise it can block indefinitely, for example if the Subscriber is
        # not reachable.
        # If messages are pending on the socket, we wait 100ms for them to be
        # sent then we discard them.
        if hasattr(self, "socket"):
            self.close(linger=100)


class ThreadController:
    """A helper class used by queue subscribers to control the background
    thread.  This is usually created and started in one go by
    :meth:`~logbook.queues.ZeroMQSubscriber.dispatch_in_background` or
    a comparable function.
    """

    def __init__(self, subscriber, setup=None):
        self.setup = setup
        self.subscriber = subscriber
        self.running = False
        self._thread = None

    def start(self):
        """Starts the task thread."""
        self.running = True
        self._thread = Thread(target=self._target)
        self._thread.daemon = True
        self._thread.start()

    def stop(self):
        """Stops the task thread."""
        if self.running:
            self.running = False
            self._thread.join()
            self._thread = None

    def _target(self):
        if self.setup is not None:
            self.setup.push_context()
        try:
            while self.running:
                self.subscriber.dispatch_once(timeout=0.05)
        finally:
            if self.setup is not None:
                self.setup.pop_context()


class SubscriberBase:
    """Baseclass for all subscribers."""

    def recv(self, timeout=None):
        """Receives a single record from the socket.  Timeout of 0 means
        nonblocking, `None` means blocking and otherwise it's a timeout in
        seconds after which the function just returns with `None`.

        Subclasses have to override this.
        """
        raise NotImplementedError()

    def dispatch_once(self, timeout=None):
        """Receives one record from the socket, loads it and dispatches it.  Returns
        `True` if something was dispatched or `False` if it timed out.
        """
        rv = self.recv(timeout)
        if rv is not None:
            dispatch_record(rv)
            return True
        return False

    def dispatch_forever(self):
        """Starts a loop that dispatches log records forever."""
        while 1:
            self.dispatch_once()

    def dispatch_in_background(self, setup=None):
        """Starts a new daemonized thread that dispatches in the background.
        An optional handler setup can be provided that pushed to the new
        thread (can be any :class:`logbook.base.StackedObject`).

        Returns a :class:`ThreadController` object for shutting down
        the background thread.  The background thread will already be
        running when this function returns.
        """
        controller = ThreadController(self, setup)
        controller.start()
        return controller


class MessageQueueSubscriber(SubscriberBase):
    """A helper that acts as a message queue subscriber and will dispatch
    received log records to the active handler setup. There are multiple ways
    to use this class.

    It can be used to receive log records from a queue::

        subscriber = MessageQueueSubscriber("mongodb://localhost:27017/logging")
        record = subscriber.recv()

    But it can also be used to receive and dispatch these in one go::

        with target_handler:
            subscriber = MessageQueueSubscriber("mongodb://localhost:27017/logging")
            subscriber.dispatch_forever()

    This will take all the log records from that queue and dispatch them
    over to `target_handler`.  If you want you can also do that in the
    background::

        subscriber = MessageQueueSubscriber("mongodb://localhost:27017/logging")
        controller = subscriber.dispatch_in_background(target_handler)

    The controller returned can be used to shut down the background
    thread::

        controller.stop()
    """

    def __init__(self, uri=None, queue="logging"):
        try:
            import kombu
        except ImportError:
            raise RuntimeError("The kombu library is required.")
        if uri:
            connection = kombu.Connection(uri)

        self.queue = connection.SimpleQueue(queue)

    def __del__(self):
        try:
            self.close()
        except AttributeError:
            # subscriber partially created
            pass

    def close(self):
        self.queue.close()

    def recv(self, timeout=None):
        """Receives a single record from the socket.  Timeout of 0 means
        nonblocking, `None` means blocking and otherwise it's a timeout in
        seconds after which the function just returns with `None`.
        """
        if timeout == 0:
            try:
                rv = self.queue.get(block=False)
            except Exception:
                return
        else:
            rv = self.queue.get(timeout=timeout)

        log_record = rv.payload
        rv.ack()

        return LogRecord.from_dict(log_record)


RabbitMQSubscriber = MessageQueueSubscriber


class ZeroMQSubscriber(SubscriberBase):
    """A helper that acts as ZeroMQ subscriber and will dispatch received
    log records to the active handler setup.  There are multiple ways to
    use this class.

    It can be used to receive log records from a queue::

        subscriber = ZeroMQSubscriber("tcp://127.0.0.1:5000")
        record = subscriber.recv()

    But it can also be used to receive and dispatch these in one go::

        with target_handler:
            subscriber = ZeroMQSubscriber("tcp://127.0.0.1:5000")
            subscriber.dispatch_forever()

    This will take all the log records from that queue and dispatch them
    over to `target_handler`.  If you want you can also do that in the
    background::

        subscriber = ZeroMQSubscriber("tcp://127.0.0.1:5000")
        controller = subscriber.dispatch_in_background(target_handler)

    The controller returned can be used to shut down the background
    thread::

        controller.stop()

    If `multi` is set to `True`, the subscriber will use a `PULL` socket
    and listen to records published by a `PUSH` socket (usually via a
    :class:`ZeroMQHandler` with `multi` set to `True`). This allows a
    single subscriber to dispatch multiple handlers.
    """

    def __init__(self, uri=None, context=None, multi=False):
        try:
            import zmq
        except ImportError:
            raise RuntimeError(
                "The pyzmq library is required for the ZeroMQSubscriber."
            )
        self._zmq = zmq

        #: the zero mq context
        self.context = context or zmq.Context()

        if multi:
            #: the zero mq socket.
            self.socket = self.context.socket(zmq.PULL)
            if uri is not None:
                self.socket.bind(uri)
        else:
            #: the zero mq socket.
            self.socket = self.context.socket(zmq.SUB)
            if uri is not None:
                self.socket.connect(uri)
            self.socket.setsockopt_unicode(zmq.SUBSCRIBE, "")

    def __del__(self):
        try:
            self.close()
        except AttributeError:
            # subscriber partially created
            pass

    def close(self):
        """Closes the zero mq socket."""
        self.socket.close()

    def recv(self, timeout=None):
        """Receives a single record from the socket.  Timeout of 0 means
        nonblocking, `None` means blocking and otherwise it's a timeout in
        seconds after which the function just returns with `None`.
        """
        if timeout is None:
            rv = self.socket.recv()
        elif not timeout:
            rv = self.socket.recv(self._zmq.NOBLOCK)
            if rv is None:
                return
        else:
            if not self._zmq.select([self.socket], [], [], timeout)[0]:
                return
            rv = self.socket.recv(self._zmq.NOBLOCK)
        rv = rv.decode("utf-8")
        return LogRecord.from_dict(json.loads(rv))


def _fix_261_mplog():
    """necessary for older python's to disable a broken monkeypatch
    in the logging module.  See multiprocessing/util.py for the
    hasattr() check.  At least in Python 2.6.1 the multiprocessing
    module is not imported by logging and as such the test in
    the util fails.
    """
    import logging
    import multiprocessing

    logging.multiprocessing = multiprocessing


class MultiProcessingHandler(Handler):
    """Implements a handler that dispatches over a queue to a different
    process.  It is connected to a subscriber with a
    :class:`multiprocessing.Queue`::

        from multiprocessing import Queue
        from logbook.queues import MultiProcessingHandler

        queue = Queue(-1)
        handler = MultiProcessingHandler(queue)

    """

    def __init__(self, queue, level=NOTSET, filter=None, bubble=False):
        Handler.__init__(self, level, filter, bubble)
        self.queue = queue
        _fix_261_mplog()

    def emit(self, record):
        self.queue.put_nowait(record.to_dict(json_safe=True))


class MultiProcessingSubscriber(SubscriberBase):
    """Receives log records from the given multiprocessing queue and
    dispatches them to the active handler setup.  Make sure to use the same
    queue for both handler and subscriber.  Idaelly the queue is set
    up with maximum size (``-1``)::

        from multiprocessing import Queue

        queue = Queue(-1)

    It can be used to receive log records from a queue::

        subscriber = MultiProcessingSubscriber(queue)
        record = subscriber.recv()

    But it can also be used to receive and dispatch these in one go::

        with target_handler:
            subscriber = MultiProcessingSubscriber(queue)
            subscriber.dispatch_forever()

    This will take all the log records from that queue and dispatch them
    over to `target_handler`.  If you want you can also do that in the
    background::

        subscriber = MultiProcessingSubscriber(queue)
        controller = subscriber.dispatch_in_background(target_handler)

    The controller returned can be used to shut down the background
    thread::

        controller.stop()

    If no queue is provided the subscriber will create one.  This one can the
    be used by handlers::

        subscriber = MultiProcessingSubscriber()
        handler = MultiProcessingHandler(subscriber.queue)
    """

    def __init__(self, queue=None):
        if queue is None:
            from multiprocessing import Queue

            queue = Queue(-1)
        self.queue = queue
        _fix_261_mplog()

    def recv(self, timeout=None):
        if timeout is None:
            rv = self.queue.get()
        else:
            try:
                rv = self.queue.get(block=True, timeout=timeout)
            except Empty:
                return None
        return LogRecord.from_dict(rv)


class ExecnetChannelHandler(Handler):
    """Implements a handler that dispatches over a execnet channel
    to a different process.
    """

    def __init__(self, channel, level=NOTSET, filter=None, bubble=False):
        Handler.__init__(self, level, filter, bubble)
        self.channel = channel

    def emit(self, record):
        self.channel.send(record.to_dict(json_safe=True))


class ExecnetChannelSubscriber(SubscriberBase):
    """subscribes to a execnet channel"""

    def __init__(self, channel):
        self.channel = channel

    def recv(self, timeout=None):
        try:
            rv = self.channel.receive(timeout=timeout)
        except self.channel.RemoteError:
            # XXX: handle
            return None
        except (self.channel.TimeoutError, EOFError):
            return None
        else:
            return LogRecord.from_dict(rv)


class TWHThreadController:
    """A very basic thread controller that pulls things in from a
    queue and sends it to a handler.  Both queue and handler are
    taken from the passed :class:`ThreadedWrapperHandler`.
    """

    class Command:
        stop = object()
        emit = object()
        emit_batch = object()

    def __init__(self, wrapper_handler):
        self.wrapper_handler = wrapper_handler
        self.running = False
        self._thread = None

    def start(self):
        """Starts the task thread."""
        self.running = True
        self._thread = Thread(target=self._target)
        self._thread.daemon = True
        self._thread.start()

    def stop(self):
        """Stops the task thread."""
        if self.running:
            self.wrapper_handler.queue.put_nowait((self.Command.stop,))
            self._thread.join()
            self._thread = None

    def _target(self):
        while 1:
            item = self.wrapper_handler.queue.get()
            command, data = item[0], item[1:]
            if command is self.Command.stop:
                self.running = False
                break
            elif command is self.Command.emit:
                (record,) = data
                self.wrapper_handler.handler.emit(record)
            elif command is self.Command.emit_batch:
                record, reason = data
                self.wrapper_handler.handler.emit_batch(record, reason)


class ThreadedWrapperHandler(WrapperHandler):
    """This handled uses a single background thread to dispatch log records
    to a specific other handler using an internal queue.  The idea is that if
    you are using a handler that requires some time to hand off the log records
    (such as the mail handler) and would block your request, you can let
    Logbook do that in a background thread.

    The threaded wrapper handler will automatically adopt the methods and
    properties of the wrapped handler.  All the values will be reflected:

    >>> twh = ThreadedWrapperHandler(TestHandler())
    >>> from logbook import WARNING
    >>> twh.level_name = 'WARNING'
    >>> twh.handler.level_name
    'WARNING'
    """

    _direct_attrs = frozenset(["handler", "queue", "controller"])

    def __init__(self, handler, maxsize=0):
        WrapperHandler.__init__(self, handler)
        self.queue = ThreadQueue(maxsize)
        self.controller = TWHThreadController(self)
        self.controller.start()

    def close(self):
        self.controller.stop()
        self.handler.close()

    def emit(self, record):
        item = (TWHThreadController.Command.emit, record)
        try:
            self.queue.put_nowait(item)
        except Full:
            # silently drop
            pass

    def emit_batch(self, records, reason):
        item = (TWHThreadController.Command.emit_batch, records, reason)
        try:
            self.queue.put_nowait(item)
        except Full:
            # silently drop
            pass


class GroupMember(ThreadController):
    def __init__(self, subscriber, queue):
        ThreadController.__init__(self, subscriber, None)
        self.queue = queue

    def _target(self):
        if self.setup is not None:
            self.setup.push_context()
        try:
            while self.running:
                if record := self.subscriber.recv():
                    try:
                        self.queue.put(record, timeout=0.05)
                    except Full:
                        pass
        finally:
            if self.setup is not None:
                self.setup.pop_context()


class SubscriberGroup(SubscriberBase):
    """This is a subscriber which represents a group of subscribers.

    This is helpful if you are writing a server-like application which has
    "slaves". This way a user is easily able to view every log record which
    happened somewhere in the entire system without having to check every
    single slave::

        subscribers = SubscriberGroup(
            [MultiProcessingSubscriber(queue), ZeroMQSubscriber("tcp://127.0.0.1:5000")]
        )
        with target_handler:
            subscribers.dispatch_forever()
    """

    def __init__(self, subscribers=None, queue_limit=10):
        self.members = []
        self.queue = ThreadQueue(queue_limit)
        for subscriber in subscribers or []:
            self.add(subscriber)

    def add(self, subscriber):
        """Adds the given `subscriber` to the group."""
        member = GroupMember(subscriber, self.queue)
        member.start()
        self.members.append(member)

    def recv(self, timeout=None):
        try:
            return self.queue.get(timeout=timeout)
        except Empty:
            return

    def stop(self):
        """Stops the group from internally recieving any more messages, once the
        internal queue is exhausted :meth:`recv` will always return `None`.
        """
        for member in self.members:
            member.stop()



"""
logbook._fallback
~~~~~~~~~~~~~~~~~

Fallback implementations in case speedups is not around.

:copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import threading
from collections.abc import Callable, Iterator
from contextvars import ContextVar
from itertools import chain, count
from typing import Any, Generic, TypeVar

from logbook.helpers import get_iterator_next_method

_missing = object()

T = TypeVar("T")


class group_reflected_property:
    def __init__(self, default, *, fallback=_missing):
        self.default = default
        self.fallback = fallback
        self.prop_name = None
        self.attr_name = None

    def __set_name__(self, owner: type, name: str) -> None:
        self.prop_name = name
        self.attr_name = f"_{name}"

    def __get__(self, instance: Any, owner: type) -> Any:
        if instance is None:
            return self
        if self.attr_name is None:
            raise TypeError("property is not bound to a class")
        rv = getattr(instance, self.attr_name, _missing)
        if rv is not _missing and rv != self.fallback:
            return rv
        if instance.group is None:
            return self.default
        return getattr(instance.group, self.prop_name)

    def __set__(self, instance: Any, value: Any) -> None:
        setattr(instance, self.attr_name, value)

    def __delete__(self, instance: Any) -> None:
        delattr(instance, self.attr_name)


class ApplicationBound:
    def __init__(self, obj):
        self.__obj = obj

    def __enter__(self):
        self.__obj.push_application()
        return self.__obj

    def __exit__(self, exc_type, exc_value, tb):
        self.__obj.pop_application()


class StackedObject:
    """Baseclass for all objects that provide stack manipulation
    operations.
    """

    def push_context(self):
        """Pushes the stacked object to the context stack."""
        raise NotImplementedError()

    def pop_context(self):
        """Pops the stacked object from the context stack."""
        raise NotImplementedError()

    def push_application(self):
        """Pushes the stacked object to the application stack."""
        raise NotImplementedError()

    def pop_application(self):
        """Pops the stacked object from the application stack."""
        raise NotImplementedError()

    def __enter__(self):
        self.push_context()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.pop_context()

    def applicationbound(self):
        """Can be used in combination with the `with` statement to
        execute code while the object is bound to the application.
        """
        return ApplicationBound(self)


class _StackState(Generic[T]):
    """The context stack, stored as a linked list.

    Each node holds one pushed object and points at the node below it.
    Pushing creates a new node; popping goes back to the parent. Because
    popping returns to the same node object rather than rebuilding the
    stack, a merge result cached on that node (``merged``) is still there
    if the context pops back to it later. The empty stack is a single
    root node with ``item = parent = None``.
    """

    __slots__ = ("item", "merged", "parent", "size")

    def __init__(
        self, item: tuple[int, T] | None, parent: _StackState[T] | None
    ) -> None:
        self.item = item
        self.parent = parent
        self.size = 0 if parent is None else parent.size + 1
        self.merged: tuple[tuple[tuple[int, T], ...], tuple[T, ...]] | None = None

    def __iter__(self) -> Iterator[tuple[int, T]]:
        items = []
        node = self
        while node.parent is not None:
            assert node.item is not None
            items.append(node.item)
            node = node.parent
        return iter(reversed(items))

    def __repr__(self) -> str:
        return f"<_StackState len={self.size}>"


class ContextStackManager(Generic[T]):
    """Helper class for context objects that manages a stack of
    objects.
    """

    def __init__(self) -> None:
        self._write_lock = threading.Lock()
        # iter_context_objects compares globals with "is". Python reuses
        # one object for every (), so if the stack empties out again later,
        # an old cache built against the earlier empty stack still matches.
        # That is fine: both are empty, so the cached merge is the same.
        self._global: tuple[tuple[int, T], ...] = ()
        # Every context starts from this one root node. Sharing it across
        # threads is safe: nodes never change after creation, except for
        # "merged", which is written in a single assignment.
        self._root: _StackState[T] = _StackState(None, None)
        self._context_stack: ContextVar[_StackState[T]] = ContextVar(
            "stack", default=self._root
        )
        self._stackop: Callable[[], int] = get_iterator_next_method(count())

    def iter_context_objects(self) -> Iterator[T]:
        """Returns an iterator over all objects for the combined
        application and context cache.
        """
        node = self._context_stack.get()
        current_global = self._global

        memo = node.merged
        if memo is not None and memo[0] is current_global:
            return iter(memo[1])

        stack_objects = sorted(chain(current_global, node), reverse=True)
        objects = tuple(x[1] for x in stack_objects)
        node.merged = (current_global, objects)

        # Each node below keeps its own cache, ready for when the context
        # pops back to that depth. But a cache built against an old global
        # stack will never be used again, and it keeps that stack's handlers
        # alive. Clear those out now rather than waiting for the context to
        # unwind. If another thread stores a fresh cache at the same time,
        # either outcome is fine; the worst case is one extra recompute.
        ancestor = node.parent
        while ancestor is not None:
            memo = ancestor.merged
            if memo is not None and memo[0] is not current_global:
                ancestor.merged = None
            ancestor = ancestor.parent

        return iter(objects)

    def push_context(self, obj: T) -> None:
        node = self._context_stack.get()
        self._context_stack.set(_StackState((self._stackop(), obj), node))

    def pop_context(self) -> T:
        node = self._context_stack.get()
        if node.parent is None:
            raise AssertionError("no objects on stack")
        assert node.item is not None
        self._context_stack.set(node.parent)
        return node.item[1]

    def push_application(self, obj: T) -> None:
        with self._write_lock:
            item = (self._stackop(), obj)
            self._global = (*self._global, item)
            # Not needed for correctness (the "is" check would reject it),
            # but the root node lives as long as the manager, and its cache
            # holds the old global stack alive. Clear it now instead of
            # waiting for an iteration on an empty context stack, which may
            # never happen.
            self._root.merged = None

    def pop_application(self) -> T:
        with self._write_lock:
            current = self._global
            if not current:
                raise AssertionError("no objects on application stack")
            self._global = current[:-1]
            # See push_application. Here it matters more: without this,
            # the root's cache would keep the popped handler alive.
            self._root.merged = None
            return current[-1][1]



"""
logbook.handlers
~~~~~~~~~~~~~~~~

The handler interface and builtin handlers.

:copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
:license: BSD, see LICENSE for more details.
"""

import errno
import gzip
import math
import os
import re
import socket
import ssl
import stat
import sys
import traceback
import warnings
from collections import deque
from collections.abc import Mapping
from datetime import datetime, timedelta
from hashlib import sha1
from textwrap import dedent

from typing_extensions import deprecated

from logbook.base import (
    CRITICAL,
    DEBUG,
    ERROR,
    INFO,
    NOTICE,
    NOTSET,
    TRACE,
    WARNING,
    ContextObject,
    ContextStackManager,
    Flags,
    _datetime_factory,
    _missing,
    level_name_property,
    lookup_level,
)
from logbook.concurrency import _new_fine_grained_lock
from logbook.helpers import datetime_utcnow, rename

DEFAULT_FORMAT_STRING = (
    "[{record.time:%Y-%m-%d %H:%M:%S.%f%z}] "
    "{record.level_name}: {record.channel}: {record.message}"
)

SYSLOG_FORMAT_STRING = "{record.channel}: {record.message}"
NTLOG_FORMAT_STRING = dedent(
    """
    Message Level: {record.level_name}
    Location: {record.filename}:{record.lineno}
    Module: {record.module}
    Function: {record.func_name}
    Exact Time: {record.time:%Y-%m-%d %H:%M:%S}

    Event provided Message:

    {record.message}
    """
).lstrip()

TEST_FORMAT_STRING = "[{record.level_name}] {record.channel}: {record.message}"
MAIL_FORMAT_STRING = dedent(
    """
    Subject: {handler.subject}

    Message type:       {record.level_name}
    Location:           {record.filename}:{record.lineno}
    Module:             {record.module}
    Function:           {record.func_name}
    Time:               {record.time:%Y-%m-%d %H:%M:%S}

    Message:

    {record.message}
    """
).lstrip()

MAIL_RELATED_FORMAT_STRING = dedent(
    """
    Message type:       {record.level_name}
    Location:           {record.filename}:{record.lineno}
    Module:             {record.module}
    Function:           {record.func_name}
    {record.message}
    """
).lstrip()

SYSLOG_PORT = 514

REGTYPE = type(re.compile("I'm a regular expression!"))


def create_syshandler(application_name, level=NOTSET):
    """Creates the handler the operating system provides.  On Unix systems
    this creates a :class:`SyslogHandler`, on Windows sytems it will
    create a :class:`NTEventLogHandler`.
    """
    if os.name == "nt":
        return NTEventLogHandler(application_name, level=level)
    return SyslogHandler(application_name, level=level)


class _HandlerType(type):
    """The metaclass of handlers injects a destructor if the class has an
    overridden close method.  This makes it possible that the default
    handler class as well as all subclasses that don't need cleanup to be
    collected with less overhead.
    """

    def __new__(cls, name, bases, d):
        # aha, that thing has a custom close method.  We will need a magic
        # __del__ for it to be called on cleanup.
        if (
            bases != (ContextObject,)
            and "close" in d
            and "__del__" not in d
            and not any(hasattr(x, "__del__") for x in bases)
        ):

            def _magic_del(self):
                try:
                    self.close()
                except Exception:
                    # del is also invoked when init fails, so we better just
                    # ignore any exception that might be raised here
                    pass

            d["__del__"] = _magic_del
        return type.__new__(cls, name, bases, d)


class Handler(ContextObject, metaclass=_HandlerType):
    """Handler instances dispatch logging events to specific destinations.

    The base handler class. Acts as a placeholder which defines the Handler
    interface. Handlers can optionally use Formatter instances to format
    records as desired. By default, no formatter is specified; in this case,
    the 'raw' message as determined by record.message is logged.

    To bind a handler you can use the :meth:`push_application` or
    :meth:`push_context` methods.  This will push the handler on a stack of
    handlers.  To undo this, use the :meth:`pop_application` and
    :meth:`pop_context` methods::

        handler = MyHandler()
        handler.push_application()
        # all here goes to that handler
        handler.pop_application()

    By default messages sent to that handler will not go to a handler on
    an outer level on the stack, if handled.  This can be changed by
    setting bubbling to `True`.

    There is also a context manager to setup the handler for the duration
    of a `with`-block::

        with handler.applicationbound():
            ...

    Because binding to the context is a common operation, it is the
    default::

        with handler:
            ...
    """

    stack_manager = ContextStackManager()

    #: a flag for this handler that can be set to `True` for handlers that
    #: are consuming log records but are not actually displaying it.  This
    #: flag is set for the :class:`NullHandler` for instance.
    blackhole = False

    def __init__(self, level=NOTSET, filter=None, bubble=False):
        #: the level for the handler.  Defaults to `NOTSET` which
        #: consumes all entries.
        self.level = lookup_level(level)
        #: the formatter to be used on records.  This is a function
        #: that is passed a log record as first argument and the
        #: handler as second and returns something formatted
        #: (usually a unicode string)
        self.formatter = None
        #: the filter to be used with this handler
        self.filter = filter
        #: the bubble flag of this handler
        self.bubble = bubble

    level_name = level_name_property()

    def format(self, record):
        """Formats a record with the given formatter.  If no formatter
        is set, the record message is returned.  Generally speaking the
        return value is most likely a unicode string, but nothing in
        the handler interface requires a formatter to return a unicode
        string.

        The combination of a handler and formatter might have the
        formatter return an XML element tree for example.
        """
        if self.formatter is None:
            return record.message
        return self.formatter(record, self)

    def should_handle(self, record):
        """Returns `True` if this handler wants to handle the record.  The
        default implementation checks the level.
        """
        return record.level >= self.level

    def handle(self, record):
        """Emits the record and falls back.  It tries to :meth:`emit` the
        record and if that fails, it will call into :meth:`handle_error` with
        the record and traceback.  This function itself will always emit
        when called, even if the logger level is higher than the record's
        level.

        If this method returns `False` it signals to the calling function that
        no recording took place in which case it will automatically bubble.
        This should not be used to signal error situations.  The default
        implementation always returns `True`.
        """
        try:
            self.emit(record)
        except Exception:
            self.handle_error(record, sys.exc_info())
        return True

    def emit(self, record):
        """Emit the specified logging record.  This should take the
        record and deliver it to whereever the handler sends formatted
        log records.
        """

    def emit_batch(self, records, reason):
        """Some handlers may internally queue up records and want to forward
        them at once to another handler.  For example the
        :class:`~logbook.FingersCrossedHandler` internally buffers
        records until a level threshold is reached in which case the buffer
        is sent to this method and not :meth:`emit` for each record.

        The default behaviour is to call :meth:`emit` for each record in
        the buffer, but handlers can use this to optimize log handling.  For
        instance the mail handler will try to batch up items into one mail
        and not to emit mails for each record in the buffer.

        Note that unlike :meth:`emit` there is no wrapper method like
        :meth:`handle` that does error handling.  The reason is that this
        is intended to be used by other handlers which are already protected
        against internal breakage.

        `reason` is a string that specifies the rason why :meth:`emit_batch`
        was called, and not :meth:`emit`.  The following are valid values:

        ``'buffer'``
            Records were buffered for performance reasons or because the
            records were sent to another process and buffering was the only
            possible way.  For most handlers this should be equivalent to
            calling :meth:`emit` for each record.

        ``'escalation'``
            Escalation means that records were buffered in case the threshold
            was exceeded.  In this case, the last record in the iterable is the
            record that triggered the call.

        ``'group'``
            All the records in the iterable belong to the same logical
            component and happened in the same process.  For example there was
            a long running computation and the handler is invoked with a bunch
            of records that happened there.  This is similar to the escalation
            reason, just that the first one is the significant one, not the
            last.

        If a subclass overrides this and does not want to handle a specific
        reason it must call into the superclass because more reasons might
        appear in future releases.

        Example implementation::

            def emit_batch(self, records, reason):
                if reason not in ("escalation", "group"):
                    Handler.emit_batch(self, records, reason)
                ...
        """
        for record in records:
            self.emit(record)

    def close(self):
        """Tidy up any resources used by the handler.  This is automatically
        called by the destructor of the class as well, but explicit calls are
        encouraged.  Make sure that multiple calls to close are possible.
        """

    def handle_error(self, record, exc_info):
        """Handle errors which occur during an emit() call.  The behaviour of
        this function depends on the current `errors` setting.

        Check :class:`Flags` for more information.
        """
        try:
            behaviour = Flags.get_flag("errors", "print")
            if behaviour == "raise":
                raise exc_info[1]
            elif behaviour == "print":
                traceback.print_exception(*exc_info, file=sys.stderr)
                sys.stderr.write(
                    f"Logged from file {record.filename}, line {record.lineno}\n"
                )
        except OSError:
            pass


class NullHandler(Handler):
    """A handler that does nothing.

    Useful to silence logs above a certain location in the handler stack::

        handler = NullHandler()
        handler.push_application()

    NullHandlers swallow all logs sent to them, and do not bubble them onwards.

    """

    blackhole = True

    def __init__(self, level=NOTSET, filter=None):
        super().__init__(level=level, filter=filter, bubble=False)


class WrapperHandler(Handler):
    """A class that can wrap another handler and redirect all calls to the
    wrapped handler::

        handler = WrapperHandler(other_handler)

    Subclasses should override the :attr:`_direct_attrs` attribute as
    necessary.
    """

    #: a set of direct attributes that are not forwarded to the inner
    #: handler.  This has to be extended as necessary.
    _direct_attrs = frozenset(["handler"])

    def __init__(self, handler):
        self.handler = handler

    def __getattr__(self, name):
        return getattr(self.handler, name)

    def __setattr__(self, name, value):
        if name in self._direct_attrs:
            return Handler.__setattr__(self, name, value)
        setattr(self.handler, name, value)


class StringFormatter:
    """Many handlers format the log entries to text format.  This is done
    by a callable that is passed a log record and returns an unicode
    string.  The default formatter for this is implemented as a class so
    that it becomes possible to hook into every aspect of the formatting
    process.
    """

    def __init__(self, format_string):
        self.format_string = format_string

    @property
    def format_string(self):
        return self._format_string

    @format_string.setter
    def format_string(self, value):
        self._format_string = value
        self._formatter = value

    def format_record(self, record, handler):
        try:
            return self._formatter.format(record=record, handler=handler)
        except UnicodeEncodeError:
            # self._formatter is a str, but some of the record items
            # are unicode
            fmt = self._formatter.decode("ascii", "replace")
            return fmt.format(record=record, handler=handler)
        except UnicodeDecodeError:
            # self._formatter is unicode, but some of the record items
            # are non-ascii str
            fmt = self._formatter.encode("ascii", "replace")
            return fmt.format(record=record, handler=handler)

    def format_exception(self, record):
        return record.formatted_exception

    def __call__(self, record, handler):
        line = self.format_record(record, handler)
        if exc := self.format_exception(record):
            line += "\n" + exc
        return line


class StringFormatterHandlerMixin:
    """A mixin for handlers that provides a default integration for the
    :class:`~logbook.StringFormatter` class.  This is used for all handlers
    by default that log text to a destination.
    """

    #: a class attribute for the default format string to use if the
    #: constructor was invoked with `None`.
    default_format_string = DEFAULT_FORMAT_STRING

    #: the class to be used for string formatting
    formatter_class = StringFormatter

    def __init__(self, format_string):
        if format_string is None:
            format_string = self.default_format_string

        #: the currently attached format string as new-style format
        #: string.
        self.format_string = format_string

    @property
    def format_string(self):
        if isinstance(self.formatter, StringFormatter):
            return self.formatter.format_string

    @format_string.setter
    def format_string(self, value):
        if value is None:
            self.formatter = None
        else:
            self.formatter = self.formatter_class(value)


class HashingHandlerMixin:
    """Mixin class for handlers that are hashing records."""

    def hash_record_raw(self, record):
        """Returns a hashlib object with the hash of the record."""
        hash = sha1()
        hash.update(("%d\x00" % record.level).encode("ascii"))  # noqa: UP031
        hash.update((record.channel or "").encode("utf-8") + b"\x00")
        hash.update(record.filename.encode("utf-8") + b"\x00")
        hash.update(str(record.lineno).encode("utf-8"))
        return hash

    def hash_record(self, record):
        """Returns a hash for a record to keep it apart from other records.
        This is used for the `record_limit` feature.  By default
        The level, channel, filename and location are hashed.

        Calls into :meth:`hash_record_raw`.
        """
        return self.hash_record_raw(record).hexdigest()


class LimitingHandlerMixin(HashingHandlerMixin):
    """Mixin class for handlers that want to limit emitting records.

    In the default setting it delivers all log records but it can be set up
    to not send more than n mails for the same record each hour to not
    overload an inbox and the network in case a message is triggered multiple
    times a minute.  The following example limits it to 60 mails an hour::

        from datetime import timedelta

        handler = MailHandler(record_limit=1, record_delta=timedelta(minutes=1))
    """

    def __init__(self, record_limit, record_delta):
        self.record_limit = record_limit
        self._limit_lock = _new_fine_grained_lock()
        self._record_limits = {}
        if record_delta is None:
            record_delta = timedelta(seconds=60)
        elif isinstance(record_delta, (int, float)):
            record_delta = timedelta(seconds=record_delta)
        self.record_delta = record_delta

    def check_delivery(self, record):
        """Helper function to check if data should be delivered by this
        handler.  It returns a tuple in the form ``(suppression_count,
        allow)``.  The first one is the number of items that were not delivered
        so far, the second is a boolean flag if a delivery should happen now.
        """
        if self.record_limit is None:
            return 0, True
        hash = self.hash_record(record)
        self._limit_lock.acquire()
        try:
            allow_delivery = None
            suppression_count = old_count = 0
            first_count = now = datetime_utcnow()

            if hash in self._record_limits:
                last_count, suppression_count = self._record_limits[hash]
                if last_count + self.record_delta < now:
                    allow_delivery = True
                else:
                    first_count = last_count
                    old_count = suppression_count

            if (
                not suppression_count
                and len(self._record_limits) >= self.max_record_cache
            ):
                cache_items = sorted(self._record_limits.items())
                del cache_items[: int(self._record_limits) * self.record_cache_prune]
                self._record_limits = dict(cache_items)

            self._record_limits[hash] = (first_count, old_count + 1)

            if allow_delivery is None:
                allow_delivery = old_count < self.record_limit
            return suppression_count, allow_delivery
        finally:
            self._limit_lock.release()


class StreamHandler(Handler, StringFormatterHandlerMixin):
    """a handler class which writes logging records, appropriately formatted,
    to a stream. note that this class does not close the stream, as sys.stdout
    or sys.stderr may be used.

    If a stream handler is used in a `with` statement directly it will
    :meth:`close` on exit to support this pattern::

        with StreamHandler(my_stream):
            pass

    .. admonition:: Notes on the encoding

       On Python 3, the encoding parameter is only used if a stream was
       passed that was opened in binary mode.
    """

    def __init__(
        self,
        stream,
        level=NOTSET,
        format_string=None,
        encoding=None,
        filter=None,
        bubble=False,
    ):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)
        self.encoding = encoding
        self.lock = _new_fine_grained_lock()
        if stream is not _missing:
            self.stream = stream

    def __enter__(self):
        return Handler.__enter__(self)

    def __exit__(self, exc_type, exc_value, tb):
        self.close()
        return Handler.__exit__(self, exc_type, exc_value, tb)

    def ensure_stream_is_open(self):
        """this method should be overriden in sub-classes to ensure that the
        inner stream is open
        """
        pass

    def close(self):
        """The default stream handler implementation is not to close
        the wrapped stream but to flush it.
        """
        self.flush()

    def flush(self):
        """Flushes the inner stream."""
        if self.stream is not None and hasattr(self.stream, "flush"):
            self.stream.flush()

    def encode(self, msg):
        """Encodes the message to the stream encoding."""
        return msg + "\n"

    def write(self, item):
        """Writes a bytestring to the stream."""
        self.stream.write(item)

    def emit(self, record):
        msg = self.format(record)
        self.lock.acquire()
        try:
            self.ensure_stream_is_open()
            self.write(self.encode(msg))
            if self.should_flush():
                self.flush()
        finally:
            self.lock.release()

    def should_flush(self):
        return True


class FileHandler(StreamHandler):
    """A handler that does the task of opening and closing files for you.
    By default the file is opened right away, but you can also `delay`
    the open to the point where the first message is written.

    This is useful when the handler is used with a
    :class:`~logbook.FingersCrossedHandler` or something similar.
    """

    def __init__(
        self,
        filename,
        mode="a",
        encoding=None,
        level=NOTSET,
        format_string=None,
        delay=False,
        filter=None,
        bubble=False,
    ):
        if encoding is None:
            encoding = "utf-8"
        StreamHandler.__init__(
            self, None, level, format_string, encoding, filter, bubble
        )
        self._filename = os.path.abspath(filename)
        self._mode = mode
        if delay:
            self.stream = None
        else:
            self._open()

    def _open(self, mode=None):
        if mode is None:
            mode = self._mode
        self.stream = open(self._filename, mode, encoding=self.encoding)

    def write(self, item):
        self.ensure_stream_is_open()
        if isinstance(item, bytes):
            self.stream.buffer.write(item)
        else:
            self.stream.write(item)

    def close(self):
        self.lock.acquire()
        try:
            if self.stream is not None:
                self.flush()
                self.stream.close()
                self.stream = None
        finally:
            self.lock.release()

    def encode(self, record):
        # encodes based on the stream settings, so the stream has to be
        # open at the time this function is called.
        self.ensure_stream_is_open()
        return StreamHandler.encode(self, record)

    def ensure_stream_is_open(self):
        if self.stream is None:
            self._open()


class GZIPCompressionHandler(FileHandler):
    def __init__(
        self,
        filename,
        encoding=None,
        level=NOTSET,
        format_string=None,
        delay=False,
        filter=None,
        bubble=False,
        compression_quality=9,
    ):
        self._compression_quality = compression_quality
        super().__init__(
            filename,
            mode="wb",
            encoding=encoding,
            level=level,
            format_string=format_string,
            delay=delay,
            filter=filter,
            bubble=bubble,
        )

    def _open(self, mode=None):
        if mode is None:
            mode = self._mode
        self.stream = gzip.open(
            self._filename, mode, compresslevel=self._compression_quality
        )

    def write(self, item):
        if isinstance(item, str):
            item = item.encode(encoding=self.encoding)
        self.ensure_stream_is_open()
        self.stream.write(item)

    def should_flush(self):
        # gzip manages writes independently. Flushing prematurely could mean
        # duplicate flushes and thus bloated files
        return False


class BrotliCompressionHandler(FileHandler):
    def __init__(
        self,
        filename,
        encoding=None,
        level=NOTSET,
        format_string=None,
        delay=False,
        filter=None,
        bubble=False,
        compression_window_size=4 * 1024**2,
        compression_quality=11,
    ):
        super().__init__(
            filename,
            mode="wb",
            encoding=encoding,
            level=level,
            format_string=format_string,
            delay=delay,
            filter=filter,
            bubble=bubble,
        )
        try:
            import brotlicffi as brotli
        except ImportError:
            try:
                import brotli
            except ImportError:
                raise RuntimeError(
                    "The brotli/brotlicffi library is required for the "
                    "BrotliCompressionHandler."
                )

        max_window_size = int(math.log(compression_window_size, 2))
        self._compressor = brotli.Compressor(
            quality=compression_quality, lgwin=max_window_size
        )

    def _open(self, mode=None):
        if mode is None:
            mode = self._mode
        self.stream = open(self._filename, mode)

    def write(self, item):
        if isinstance(item, str):
            item = item.encode(encoding=self.encoding)
        if ret := self._compressor.process(item):
            self.ensure_stream_is_open()
            self.stream.write(ret)
            super().flush()

    def should_flush(self):
        return False

    def flush(self):
        if self._compressor is not None:
            if ret := self._compressor.flush():
                self.ensure_stream_is_open()
                self.stream.write(ret)
        super().flush()

    def close(self):
        if self._compressor is not None:
            self.ensure_stream_is_open()
            self.stream.write(self._compressor.finish())
            self._compressor = None
        super().close()


class MonitoringFileHandler(FileHandler):
    """A file handler that will check if the file was moved while it was
    open.  This might happen on POSIX systems if an application like
    logrotate moves the logfile over.

    Because of different IO concepts on Windows, this handler will not
    work on a windows system.
    """

    def __init__(
        self,
        filename,
        mode="a",
        encoding="utf-8",
        level=NOTSET,
        format_string=None,
        delay=False,
        filter=None,
        bubble=False,
    ):
        FileHandler.__init__(
            self, filename, mode, encoding, level, format_string, delay, filter, bubble
        )
        if os.name == "nt":
            raise RuntimeError("MonitoringFileHandler does not support Windows")
        self._query_fd()

    def _query_fd(self):
        if self.stream is None:
            self._last_stat = None, None
        else:
            try:
                st = os.stat(self._filename)
            except OSError:
                e = sys.exc_info()[1]
                if e.errno != errno.ENOENT:
                    raise
                self._last_stat = None, None
            else:
                self._last_stat = st[stat.ST_DEV], st[stat.ST_INO]

    def emit(self, record):
        msg = self.format(record)
        self.lock.acquire()
        try:
            last_stat = self._last_stat
            self._query_fd()
            if last_stat != self._last_stat and self.stream is not None:
                self.flush()
                self.stream.close()
                self.stream = None
            self.ensure_stream_is_open()
            self.write(self.encode(msg))
            self.flush()
            self._query_fd()
        finally:
            self.lock.release()


class StderrHandler(StreamHandler):
    """A handler that writes to what is currently at stderr.  At the first
    glance this appears to just be a :class:`StreamHandler` with the stream
    set to :data:`sys.stderr` but there is a difference: if the handler is
    created globally and :data:`sys.stderr` changes later, this handler will
    point to the current `stderr`, whereas a stream handler would still
    point to the old one.
    """

    def __init__(self, level=NOTSET, format_string=None, filter=None, bubble=False):
        StreamHandler.__init__(
            self, _missing, level, format_string, None, filter, bubble
        )

    @property
    def stream(self):
        return sys.stderr


class RotatingFileHandler(FileHandler):
    """This handler rotates based on file size.  Once the maximum size
    is reached it will reopen the file and start with an empty file
    again.  The old file is moved into a backup copy (named like the
    file, but with a ``.backupnumber`` appended to the file.  So if
    you are logging to ``mail`` the first backup copy is called
    ``mail.1``.)

    The default number of backups is 5.  Unlike a similar logger from
    the logging package, the backup count is mandatory because just
    reopening the file is dangerous as it deletes the log without
    asking on rollover.
    """

    def __init__(
        self,
        filename,
        mode="a",
        encoding="utf-8",
        level=NOTSET,
        format_string=None,
        delay=False,
        max_size=1024 * 1024,
        backup_count=5,
        filter=None,
        bubble=False,
    ):
        FileHandler.__init__(
            self, filename, mode, encoding, level, format_string, delay, filter, bubble
        )
        self.max_size = max_size
        self.backup_count = backup_count
        assert backup_count > 0, "at least one backup file has to be specified"

    def should_rollover(self, record, bytes):
        self.stream.seek(0, 2)
        return self.stream.tell() + bytes >= self.max_size

    def perform_rollover(self):
        self.stream.close()
        for x in range(self.backup_count - 1, 0, -1):
            src = f"{self._filename}.{x}"
            dst = f"{self._filename}.{x + 1}"
            try:
                rename(src, dst)
            except OSError:
                e = sys.exc_info()[1]
                if e.errno != errno.ENOENT:
                    raise
        rename(self._filename, self._filename + ".1")
        self._open("w")

    def emit(self, record):
        msg = self.format(record)
        self.lock.acquire()
        try:
            msg = self.encode(msg)
            if self.should_rollover(record, len(msg)):
                self.perform_rollover()
            self.write(msg)
            self.flush()
        finally:
            self.lock.release()


class TimedRotatingFileHandler(FileHandler):
    """This handler rotates based on dates.  It will name the file
    after the filename you specify and the `date_format` pattern.

    So for example if you configure your handler like this::

        handler = TimedRotatingFileHandler("/var/log/foo.log", date_format="%Y-%m-%d")

    The filenames for the logfiles will look like this:

    .. code-block:: text

        /var/log/foo-2010-01-10.log
        /var/log/foo-2010-01-11.log
        ...

    By default it will keep all these files around, if you want to limit
    them, you can specify a `backup_count`.

    You may supply an optional `rollover_format`. This allows you to specify
    the format for the filenames of rolled-over files.
    the format as

    So for example if you configure your handler like this::

        handler = TimedRotatingFileHandler(
            "/var/log/foo.log",
            date_format="%Y-%m-%d",
            rollover_format="{basename}{ext}.{timestamp}",
        )

    The filenames for the logfiles will look like this::

    .. code-block:: text

        /var/log/foo.log.2010-01-10
        /var/log/foo.log.2010-01-11
        ...

    Finally, an optional argument `timed_filename_for_current` may be set to
    false if you wish to have the current log file match the supplied filename
    until it is rolled over
    """

    def __init__(
        self,
        filename,
        mode="a",
        encoding="utf-8",
        level=NOTSET,
        format_string=None,
        date_format="%Y-%m-%d",
        backup_count=0,
        filter=None,
        bubble=False,
        timed_filename_for_current=True,
        rollover_format="{basename}-{timestamp}{ext}",
    ):
        self.date_format = date_format
        self.backup_count = backup_count

        self.rollover_format = rollover_format

        self.original_filename = filename
        self.basename, self.ext = os.path.splitext(os.path.abspath(filename))
        self.timed_filename_for_current = timed_filename_for_current

        self._timestamp = self._get_timestamp(_datetime_factory())
        if self.timed_filename_for_current:
            filename = self.generate_timed_filename(self._timestamp)
        elif os.path.exists(filename):
            self._timestamp = self._get_timestamp(
                datetime.fromtimestamp(os.stat(filename).st_mtime)
            )

        FileHandler.__init__(
            self, filename, mode, encoding, level, format_string, True, filter, bubble
        )

    def _get_timestamp(self, datetime):
        """
        Fetches a formatted string witha timestamp of the given datetime
        """
        return datetime.strftime(self.date_format)

    def generate_timed_filename(self, timestamp):
        """
        Produces a filename that includes a timestamp in the format supplied
        to the handler at init time.
        """
        timed_filename = self.rollover_format.format(
            basename=self.basename, timestamp=timestamp, ext=self.ext
        )
        return timed_filename

    def files_to_delete(self):
        """Returns a list with the files that have to be deleted when
        a rollover occours.
        """
        directory = os.path.dirname(self._filename)
        files = []
        rollover_regex = re.compile(
            self.rollover_format.format(
                basename=re.escape(self.basename),
                timestamp=".+",
                ext=re.escape(self.ext),
            )
        )
        for filename in os.listdir(directory):
            filename = os.path.join(directory, filename)
            if rollover_regex.match(filename):
                files.append((os.path.getmtime(filename), filename))
        files.sort()
        if self.backup_count > 1:
            return files[: -self.backup_count + 1]
        else:
            return files[:]

    def perform_rollover(self, new_timestamp):
        if self.stream is not None:
            self.stream.close()

        if not self.timed_filename_for_current and os.path.exists(self._filename):
            filename = self.generate_timed_filename(self._timestamp)
            os.rename(self._filename, filename)

        if self.backup_count > 0:
            for _, filename in self.files_to_delete():
                os.remove(filename)

        if self.timed_filename_for_current:
            self._filename = self.generate_timed_filename(new_timestamp)
        self._timestamp = new_timestamp

        self._open("w")

    def emit(self, record):
        msg = self.format(record)
        self.lock.acquire()
        try:
            new_timestamp = self._get_timestamp(record.time)
            if new_timestamp != self._timestamp:
                self.perform_rollover(new_timestamp)
            self.write(self.encode(msg))
            self.flush()
        finally:
            self.lock.release()


class TestHandler(Handler, StringFormatterHandlerMixin):
    """Like a stream handler but keeps the values in memory.  This
    logger provides some ways to test for the records in memory.

    Example usage::

        def my_test():
            with logbook.TestHandler() as handler:
                logger.warning("A warning")
                assert logger.has_warning("A warning")
                ...
    """

    default_format_string = TEST_FORMAT_STRING

    def __init__(
        self,
        level=NOTSET,
        format_string=None,
        filter=None,
        bubble=False,
        force_heavy_init=False,
    ):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)
        #: captures the :class:`LogRecord`\s as instances
        self.records = []
        self._formatted_records = []
        self._formatted_record_cache = []
        self._force_heavy_init = force_heavy_init

    def close(self):
        """Close all records down when the handler is closed."""
        for record in self.records:
            record.close()

    def emit(self, record):
        # keep records open because we will want to examine them after the
        # call to the emit function.  If we don't do that, the traceback
        # attribute and other things will already be removed.
        record.keep_open = True
        if self._force_heavy_init:
            record.heavy_init()
        self.records.append(record)

    @property
    def formatted_records(self):
        """Captures the formatted log records as unicode strings."""
        if len(self._formatted_record_cache) != len(self.records) or any(
            r1 != r2 for r1, r2 in zip(self.records, self._formatted_record_cache)
        ):
            self._formatted_records = [self.format(r) for r in self.records]
            self._formatted_record_cache = list(self.records)
        return self._formatted_records

    @property
    def has_criticals(self):
        """`True` if any :data:`CRITICAL` records were found."""
        return any(r.level == CRITICAL for r in self.records)

    @property
    def has_errors(self):
        """`True` if any :data:`ERROR` records were found."""
        return any(r.level == ERROR for r in self.records)

    @property
    def has_warnings(self):
        """`True` if any :data:`WARNING` records were found."""
        return any(r.level == WARNING for r in self.records)

    @property
    def has_notices(self):
        """`True` if any :data:`NOTICE` records were found."""
        return any(r.level == NOTICE for r in self.records)

    @property
    def has_infos(self):
        """`True` if any :data:`INFO` records were found."""
        return any(r.level == INFO for r in self.records)

    @property
    def has_debugs(self):
        """`True` if any :data:`DEBUG` records were found."""
        return any(r.level == DEBUG for r in self.records)

    @property
    def has_traces(self):
        """`True` if any :data:`TRACE` records were found."""
        return any(r.level == TRACE for r in self.records)

    def has_critical(self, *args, **kwargs):
        """`True` if a specific :data:`CRITICAL` log record exists.

        See :ref:`probe-log-records` for more information.
        """
        kwargs["level"] = CRITICAL
        return self._test_for(*args, **kwargs)

    def has_error(self, *args, **kwargs):
        """`True` if a specific :data:`ERROR` log record exists.

        See :ref:`probe-log-records` for more information.
        """
        kwargs["level"] = ERROR
        return self._test_for(*args, **kwargs)

    def has_warning(self, *args, **kwargs):
        """`True` if a specific :data:`WARNING` log record exists.

        See :ref:`probe-log-records` for more information.
        """
        kwargs["level"] = WARNING
        return self._test_for(*args, **kwargs)

    def has_notice(self, *args, **kwargs):
        """`True` if a specific :data:`NOTICE` log record exists.

        See :ref:`probe-log-records` for more information.
        """
        kwargs["level"] = NOTICE
        return self._test_for(*args, **kwargs)

    def has_info(self, *args, **kwargs):
        """`True` if a specific :data:`INFO` log record exists.

        See :ref:`probe-log-records` for more information.
        """
        kwargs["level"] = INFO
        return self._test_for(*args, **kwargs)

    def has_debug(self, *args, **kwargs):
        """`True` if a specific :data:`DEBUG` log record exists.

        See :ref:`probe-log-records` for more information.
        """
        kwargs["level"] = DEBUG
        return self._test_for(*args, **kwargs)

    def has_trace(self, *args, **kwargs):
        """`True` if a specific :data:`TRACE` log record exists.

        See :ref:`probe-log-records` for more information.
        """
        kwargs["level"] = TRACE
        return self._test_for(*args, **kwargs)

    def _test_for(self, message=None, channel=None, level=None):
        def _match(needle, haystack):
            """Matches both compiled regular expressions and strings"""
            if isinstance(needle, REGTYPE) and needle.search(haystack):
                return True
            if needle == haystack:
                return True
            return False

        for record in self.records:
            if level is not None and record.level != level:
                continue
            if channel is not None and record.channel != channel:
                continue
            if message is not None and not _match(message, record.message):
                continue
            return True
        return False


class MailHandler(Handler, StringFormatterHandlerMixin, LimitingHandlerMixin):
    """A handler that sends error mails.  The format string used by this
    handler are the contents of the mail plus the headers.  This is handy
    if you want to use a custom subject or ``X-`` header:

    .. blacken-docs:off

    .. code-block:: python

        handler = MailHandler(format_string='''\
        Subject: {record.level_name} on My Application

        {record.message}
        {record.extra[a_custom_injected_record]}
        ''')

    .. blacken-docs:on

    This handler will always emit text-only mails for maximum portability and
    best performance.

    In the default setting it delivers all log records but it can be set up
    to not send more than n mails for the same record each hour to not
    overload an inbox and the network in case a message is triggered multiple
    times a minute.  The following example limits it to 60 mails an hour::

        from datetime import timedelta

        handler = MailHandler(record_limit=1, record_delta=timedelta(minutes=1))

    The default timedelta is 60 seconds (one minute).

    The mail handler sends mails in a blocking manner.  If you are not
    using some centralized system for logging these messages (with the help
    of ZeroMQ or others) and the logging system slows you down you can
    wrap the handler in a :class:`logbook.queues.ThreadedWrapperHandler`
    that will then send the mails in a background thread.

    `server_addr` can be a tuple of host and port, or just a string containing
    the host to use the default port (25, or 465 if connecting securely.)

    `credentials` can be a tuple or dictionary of arguments that will be passed
    to :py:meth:`smtplib.SMTP.login`.

    `secure` should be an :class:`ssl.SSLContext` or boolean. Please read
    :ref:`ssl-security` for best practices. For backwards
    compatibility reasons, `secure` may also be a tuple or dictionary, although
    this is deprecated:

    * ``(keyfile, certfile)`` tuple
    * ``{'keyfile': keyfile, 'certfile': certfile}`` dict
    * ``()`` an empty tuple is equivalent to ``True``.

    .. versionchanged:: 0.3
       The handler supports the batching system now.

    .. versionadded:: 1.0
       `starttls` parameter added to allow disabling STARTTLS for SSL
       connections.

    .. versionchanged:: 1.0
       If `server_addr` is a string, the default port will be used.

    .. versionchanged:: 1.0
       `credentials` parameter can now be a dictionary of keyword arguments.

    .. versionchanged:: 1.0
       `secure` can now be a dictionary or boolean in addition to a tuple.

    .. versionchanged:: 1.7
        `secure` may be an :class:`ssl.SSLContext` (recommended). The tuple or
        dict form is deprecated.
    """

    default_format_string = MAIL_FORMAT_STRING
    default_related_format_string = MAIL_RELATED_FORMAT_STRING
    default_subject = "Server Error in Application"

    #: the maximum number of record hashes in the cache for the limiting
    #: feature.  Afterwards, record_cache_prune percent of the oldest
    #: entries are removed
    max_record_cache = 512

    #: the number of items to prune on a cache overflow in percent.
    record_cache_prune = 0.333

    def __init__(
        self,
        from_addr,
        recipients,
        subject=None,
        server_addr=None,
        credentials=None,
        secure=None,
        record_limit=None,
        record_delta=None,
        level=NOTSET,
        format_string=None,
        related_format_string=None,
        filter=None,
        bubble=False,
        starttls=True,
    ):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)
        LimitingHandlerMixin.__init__(self, record_limit, record_delta)
        self.from_addr = from_addr
        self.recipients = recipients
        if subject is None:
            subject = self.default_subject
        self.subject = subject
        self.server_addr = server_addr
        self.credentials = credentials
        self.secure = self._adapt_secure(secure)
        if related_format_string is None:
            related_format_string = self.default_related_format_string
        self.related_format_string = related_format_string
        self.starttls = starttls

    def _adapt_secure(self, secure):
        if secure is None or isinstance(secure, (bool, ssl.SSLContext)):
            return secure

        if isinstance(secure, tuple):
            if not secure:
                # For backwards compatibility, () translates to True
                return True
            else:
                keyfile, certfile = secure
        elif isinstance(secure, Mapping):
            keyfile = secure.get("keyfile", None)
            certfile = secure.get("certfile", None)
        else:
            raise TypeError(f"Unexpected type for `secure`: {type(secure)}")

        warnings.warn(
            "Passing keyfile and certfile are deprecated, use an SSLContext instead.",
            DeprecationWarning,
            stacklevel=3,
        )

        ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
        ctx.load_cert_chain(certfile, keyfile)

        return ctx

    @property
    def related_format_string(self):
        if isinstance(self.related_formatter, StringFormatter):
            return self.related_formatter.format_string

    @related_format_string.setter
    def related_format_string(self, value):
        if value is None:
            self.related_formatter = None
        else:
            self.related_formatter = self.formatter_class(value)

    def get_recipients(self, record):
        """Returns the recipients for a record.  By default the
        :attr:`recipients` attribute is returned for all records.
        """
        return self.recipients

    def message_from_record(self, record, suppressed):
        """Creates a new message for a record as email message object
        (:class:`email.message.Message`).  `suppressed` is the number
        of mails not sent if the `record_limit` feature is active.
        """
        from email.header import Header
        from email.message import Message

        msg = Message()
        msg.set_charset("utf-8")
        lineiter = iter(self.format(record).splitlines())
        for line in lineiter:
            if not line:
                break
            h, v = line.split(":", 1)
            # We could probably just encode everything. For the moment encode
            # only what really needed to avoid breaking a couple of tests.
            try:
                v.encode("ascii")
            except UnicodeEncodeError:
                msg[h.strip()] = Header(v.strip(), "utf-8")
            else:
                msg[h.strip()] = v.strip()

        msg.replace_header("Content-Transfer-Encoding", "8bit")

        body = "\r\n".join(lineiter)
        if suppressed:
            body += (
                f"\r\n\r\nThis message occurred additional {suppressed} "
                "time(s) and was suppressed"
            )

        msg.set_payload(body, "UTF-8")
        return msg

    def format_related_record(self, record):
        """Used for format the records that led up to another record or
        records that are related into strings.  Used by the batch formatter.
        """
        return self.related_formatter(record, self)

    def generate_mail(self, record, suppressed=0):
        """Generates the final email (:class:`email.message.Message`)
        with headers and date.  `suppressed` is the number of mails
        that were not send if the `record_limit` feature is active.
        """
        from email.utils import formatdate

        msg = self.message_from_record(record, suppressed)
        msg["From"] = self.from_addr
        msg["Date"] = formatdate()
        return msg

    def collapse_mails(self, mail, related, reason):
        """When escaling or grouped mails are"""
        if not related:
            return mail
        if reason == "group":
            title = "Other log records in the same group"
        else:
            title = "Log records that led up to this one"
        mail.set_payload(
            "{}\r\n\r\n\r\n{}:\r\n\r\n{}".format(
                mail.get_payload(),
                title,
                "\r\n\r\n".join(body.rstrip() for body in related),
            ),
            "UTF-8",
        )
        return mail

    def get_connection(self):
        """Returns an SMTP connection.  By default it reconnects for
        each sent mail.
        """
        from smtplib import SMTP, SMTP_PORT, SMTP_SSL, SMTP_SSL_PORT

        if self.secure:
            if self.starttls:
                default_port = 587
            else:
                default_port = SMTP_SSL_PORT
        else:
            default_port = SMTP_PORT

        if self.server_addr is None:
            host = "127.0.0.1"
            port = default_port
        else:
            try:
                host, port = self.server_addr
            except ValueError:
                # If server_addr is a string, the tuple unpacking will raise
                # ValueError, and we can use the default port.
                host = self.server_addr
                port = default_port

        if isinstance(self.secure, ssl.SSLContext):
            context = self.secure
        else:
            context = None

        if self.secure and not self.starttls:
            con = SMTP_SSL(host, port, context=context)
        else:
            con = SMTP(host, port)

        if self.secure and self.starttls:
            con.starttls(context=context)
            con.ehlo()

        if self.credentials is not None:
            # Allow credentials to be a tuple or dict.
            if isinstance(self.credentials, Mapping):
                credentials_args = ()
                credentials_kwargs = self.credentials
            else:
                credentials_args = self.credentials
                credentials_kwargs = dict()

            con.login(*credentials_args, **credentials_kwargs)

        return con

    def close_connection(self, con):
        """Closes the connection that was returned by
        :meth:`get_connection`.
        """
        try:
            if con is not None:
                con.quit()
        except Exception:
            pass

    def deliver(self, msg, recipients):
        """Delivers the given message to a list of recipients."""
        con = self.get_connection()
        try:
            con.sendmail(self.from_addr, recipients, msg.as_string())
        finally:
            self.close_connection(con)

    def emit(self, record):
        suppressed = 0
        if self.record_limit is not None:
            suppressed, allow_delivery = self.check_delivery(record)
            if not allow_delivery:
                return
        self.deliver(
            self.generate_mail(record, suppressed), self.get_recipients(record)
        )

    def emit_batch(self, records, reason):
        if reason not in ("escalation", "group"):
            raise RuntimeError("reason must be either 'escalation' or 'group'")
        records = list(records)
        if not records:
            return

        trigger = records.pop(-1 if reason == "escalation" else 0)
        suppressed = 0
        if self.record_limit is not None:
            suppressed, allow_delivery = self.check_delivery(trigger)
            if not allow_delivery:
                return

        trigger_mail = self.generate_mail(trigger, suppressed)
        related = [self.format_related_record(record) for record in records]

        self.deliver(
            self.collapse_mails(trigger_mail, related, reason),
            self.get_recipients(trigger),
        )


class GMailHandler(MailHandler):
    """
    A customized mail handler class for sending emails via GMail (or Google
    Apps mail)::

       handler = GMailHandler(
           "my_user@gmail.com", "mypassword", ["to_user@some_mail.com"], ...
       )  # other arguments same as MailHandler

    .. versionadded:: 0.6.0
    """

    def __init__(self, account_id, password, recipients, **kw):
        super().__init__(
            account_id,
            recipients,
            secure=True,
            server_addr=("smtp.gmail.com", 587),
            credentials=(account_id, password),
            **kw,
        )


class SyslogHandler(Handler, StringFormatterHandlerMixin):
    """A handler class which sends formatted logging records to a
    syslog server.  By default it will send to it via unix socket.
    """

    default_format_string = SYSLOG_FORMAT_STRING

    # priorities
    LOG_EMERG = 0  # system is unusable
    LOG_ALERT = 1  # action must be taken immediately
    LOG_CRIT = 2  # critical conditions
    LOG_ERR = 3  # error conditions
    LOG_WARNING = 4  # warning conditions
    LOG_NOTICE = 5  # normal but significant condition
    LOG_INFO = 6  # informational
    LOG_DEBUG = 7  # debug-level messages

    # facility codes
    LOG_KERN = 0  # kernel messages
    LOG_USER = 1  # random user-level messages
    LOG_MAIL = 2  # mail system
    LOG_DAEMON = 3  # system daemons
    LOG_AUTH = 4  # security/authorization messages
    LOG_SYSLOG = 5  # messages generated internally by syslogd
    LOG_LPR = 6  # line printer subsystem
    LOG_NEWS = 7  # network news subsystem
    LOG_UUCP = 8  # UUCP subsystem
    LOG_CRON = 9  # clock daemon
    LOG_AUTHPRIV = 10  # security/authorization messages (private)
    LOG_FTP = 11  # FTP daemon

    # other codes through 15 reserved for system use
    LOG_LOCAL0 = 16  # reserved for local use
    LOG_LOCAL1 = 17  # reserved for local use
    LOG_LOCAL2 = 18  # reserved for local use
    LOG_LOCAL3 = 19  # reserved for local use
    LOG_LOCAL4 = 20  # reserved for local use
    LOG_LOCAL5 = 21  # reserved for local use
    LOG_LOCAL6 = 22  # reserved for local use
    LOG_LOCAL7 = 23  # reserved for local use

    facility_names = {
        "auth": LOG_AUTH,
        "authpriv": LOG_AUTHPRIV,
        "cron": LOG_CRON,
        "daemon": LOG_DAEMON,
        "ftp": LOG_FTP,
        "kern": LOG_KERN,
        "lpr": LOG_LPR,
        "mail": LOG_MAIL,
        "news": LOG_NEWS,
        "syslog": LOG_SYSLOG,
        "user": LOG_USER,
        "uucp": LOG_UUCP,
        "local0": LOG_LOCAL0,
        "local1": LOG_LOCAL1,
        "local2": LOG_LOCAL2,
        "local3": LOG_LOCAL3,
        "local4": LOG_LOCAL4,
        "local5": LOG_LOCAL5,
        "local6": LOG_LOCAL6,
        "local7": LOG_LOCAL7,
    }

    level_priority_map = {
        DEBUG: LOG_DEBUG,
        INFO: LOG_INFO,
        NOTICE: LOG_NOTICE,
        WARNING: LOG_WARNING,
        ERROR: LOG_ERR,
        CRITICAL: LOG_CRIT,
    }

    def __init__(
        self,
        application_name=None,
        address=None,
        facility="user",
        socktype=socket.SOCK_DGRAM,
        level=NOTSET,
        format_string=None,
        filter=None,
        bubble=False,
        record_delimiter=None,
    ):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)
        self.application_name = application_name

        if address is None:
            if sys.platform == "darwin":
                address = "/var/run/syslog"
            else:
                address = "/dev/log"

        self.remote_address = self.address = address
        self.facility = facility
        self.socktype = socktype

        if isinstance(address, str):
            self._connect_unixsocket()
            self.enveloper = self.unix_envelope
            default_delimiter = "\x00"
        else:
            self._connect_netsocket()
            self.enveloper = self.net_envelope
            default_delimiter = "\n"

        self.record_delimiter = (
            default_delimiter if record_delimiter is None else record_delimiter
        )

        self.connection_exception = getattr(
            __builtins__, "BrokenPipeError", socket.error
        )

    def _connect_unixsocket(self):
        self.unixsocket = True
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        try:
            self.socket.connect(self.address)
        except OSError:
            self.socket.close()
            self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.socket.connect(self.address)

    def _connect_netsocket(self):
        self.unixsocket = False
        self.socket = socket.socket(socket.AF_INET, self.socktype)
        if self.socktype == socket.SOCK_STREAM:
            self.socket.connect(self.remote_address)
            self.address = self.socket.getsockname()

    def encode_priority(self, record):
        facility = self.facility_names[self.facility]
        priority = self.level_priority_map.get(record.level, self.LOG_WARNING)
        return (facility << 3) | priority

    def wrap_segments(self, record, before):
        msg = self.format(record)
        segments = [segment for segment in msg.split(self.record_delimiter)]
        return (before + segment + self.record_delimiter for segment in segments)

    def unix_envelope(self, record):
        before = "<{}>{}".format(
            self.encode_priority(record),
            self.application_name + ":" if self.application_name else "",
        )
        return self.wrap_segments(record, before)

    def net_envelope(self, record):
        # Gross but effective
        try:
            format_string = self.format_string
            application_name = self.application_name
            if (
                not application_name
                and record.channel
                and "{record.channel}: " in format_string
            ):
                self.format_string = format_string.replace("{record.channel}: ", "")
                self.application_name = record.channel
            # RFC 5424: <PRIVAL>version timestamp hostname app-name procid
            #           msgid structured-data message
            before = "<{}>1 {}Z {} {} {} - - ".format(
                self.encode_priority(record),
                record.time.isoformat(),
                socket.gethostname(),
                self.application_name if self.application_name else "-",
                record.process,
            )
            return self.wrap_segments(record, before)
        finally:
            self.format_string = format_string
            self.application_name = application_name

    def emit(self, record):
        for segment in self.enveloper(record):
            self.send_to_socket(segment.encode("utf-8"))

    def send_to_socket(self, data):
        if self.unixsocket:
            try:
                self.socket.send(data)
            except OSError:
                self._connect_unixsocket()
                self.socket.send(data)
        elif self.socktype == socket.SOCK_DGRAM:
            # the flags are no longer optional on Python 3
            self.socket.sendto(data, 0, self.address)
        else:
            try:
                self.socket.sendall(data)
            except self.connection_exception:
                self._connect_netsocket()
                self.socket.send(data)

    def close(self):
        self.socket.close()


class NTEventLogHandler(Handler, StringFormatterHandlerMixin):
    """A handler that sends to the NT event log system."""

    dllname = None
    default_format_string = NTLOG_FORMAT_STRING

    def __init__(
        self,
        application_name,
        log_type="Application",
        level=NOTSET,
        format_string=None,
        filter=None,
        bubble=False,
    ):
        Handler.__init__(self, level, filter, bubble)
        StringFormatterHandlerMixin.__init__(self, format_string)

        if os.name != "nt":
            raise RuntimeError(
                "NTLogEventLogHandler requires a Windows operating system."
            )

        try:
            import win32evtlog
            import win32evtlogutil
        except ImportError:
            raise RuntimeError(
                "The pywin32 library is required for the NTEventLogHandler."
            )

        self.application_name = application_name
        self._welu = win32evtlogutil
        dllname = self.dllname
        if not dllname:
            dllname = os.path.join(
                os.path.dirname(self._welu.__file__), "../win32service.pyd"
            )
        self.log_type = log_type
        self._welu.AddSourceToRegistry(self.application_name, dllname, log_type)

        self._default_type = win32evtlog.EVENTLOG_INFORMATION_TYPE
        self._type_map = {
            DEBUG: win32evtlog.EVENTLOG_INFORMATION_TYPE,
            INFO: win32evtlog.EVENTLOG_INFORMATION_TYPE,
            NOTICE: win32evtlog.EVENTLOG_INFORMATION_TYPE,
            WARNING: win32evtlog.EVENTLOG_WARNING_TYPE,
            ERROR: win32evtlog.EVENTLOG_ERROR_TYPE,
            CRITICAL: win32evtlog.EVENTLOG_ERROR_TYPE,
        }

    def unregister_logger(self):
        """Removes the application binding from the registry.  If you call
        this, the log viewer will no longer be able to provide any
        information about the message.
        """
        self._welu.RemoveSourceFromRegistry(self.application_name, self.log_type)

    def get_event_type(self, record):
        return self._type_map.get(record.level, self._default_type)

    def get_event_category(self, record):
        """Returns the event category for the record. Override this if you want
        to specify your own categories. This version returns 0.
        """
        return 0

    def get_message_id(self, record):
        """Returns the message ID (EventID) for the record. Override this if
        you want to specify your own ID. This version returns 1.
        """
        return 1

    def emit(self, record):
        id = self.get_message_id(record)
        cat = self.get_event_category(record)
        type = self.get_event_type(record)
        self._welu.ReportEvent(
            self.application_name, id, cat, type, [self.format(record)]
        )


class FingersCrossedHandler(Handler):
    """This handler wraps another handler and will log everything in
    memory until a certain level (`action_level`, defaults to `ERROR`)
    is exceeded.  When that happens the fingers crossed handler will
    activate forever and log all buffered records as well as records
    yet to come into another handled which was passed to the constructor.

    Alternatively it's also possible to pass a factory function to the
    constructor instead of a handler.  That factory is then called with
    the triggering log entry and the finger crossed handler to create
    a handler which is then cached.

    The idea of this handler is to enable debugging of live systems.  For
    example it might happen that code works perfectly fine 99% of the time,
    but then some exception happens.  But the error that caused the
    exception alone might not be the interesting bit, the interesting
    information were the warnings that lead to the error.

    Here a setup that enables this for a web application::

        from logbook import FileHandler
        from logbook import FingersCrossedHandler


        def issue_logging():
            def factory(record, handler):
                return FileHandler("/var/log/app/issue-%s.log" % record.time)

            return FingersCrossedHandler(factory)


        def application(environ, start_response):
            with issue_logging():
                return the_actual_wsgi_application(environ, start_response)

    Whenever an error occours, a new file in ``/var/log/app`` is created
    with all the logging calls that lead up to the error up to the point
    where the `with` block is exited.

    Please keep in mind that the :class:`~logbook.FingersCrossedHandler`
    handler is a one-time handler.  Once triggered, it will not reset.  Because
    of that you will have to re-create it whenever you bind it.  In this case
    the handler is created when it's bound to the thread.

    Due to how the handler is implemented, the filter, bubble and level
    flags of the wrapped handler are ignored.

    .. versionchanged:: 0.3

    The default behaviour is to buffer up records and then invoke another
    handler when a severity threshold was reached with the buffer emitting.
    This now enables this logger to be properly used with the
    :class:`~logbook.MailHandler`.  You will now only get one mail for
    each buffered record.  However once the threshold was reached you would
    still get a mail for each record which is why the `reset` flag was added.

    When set to `True`, the handler will instantly reset to the untriggered
    state and start buffering again::

        handler = FingersCrossedHandler(MailHandler(...), buffer_size=10, reset=True)

    .. versionadded:: 0.3
       The `reset` flag was added.
    """

    #: the reason to be used for the batch emit.  The default is
    #: ``'escalation'``.
    #:
    #: .. versionadded:: 0.3
    batch_emit_reason = "escalation"

    def __init__(
        self,
        handler,
        action_level=ERROR,
        buffer_size=0,
        pull_information=True,
        reset=False,
        filter=None,
        bubble=False,
    ):
        Handler.__init__(self, NOTSET, filter, bubble)
        self.lock = _new_fine_grained_lock()
        self._level = action_level
        if isinstance(handler, Handler):
            self._handler = handler
            self._handler_factory = None
        else:
            self._handler = None
            self._handler_factory = handler
        #: the buffered records of the handler.  Once the action is triggered
        #: (:attr:`triggered`) this list will be None.  This attribute can
        #: be helpful for the handler factory function to select a proper
        #: filename (for example time of first log record)
        self.buffered_records = deque()
        #: the maximum number of entries in the buffer.  If this is exhausted
        #: the oldest entries will be discarded to make place for new ones
        self.buffer_size = buffer_size
        self._buffer_full = False
        self._pull_information = pull_information
        self._action_triggered = False
        self._reset = reset

    def close(self):
        if self._handler is not None:
            self._handler.close()

    def enqueue(self, record):
        if self._pull_information:
            record.pull_information()
        if self._action_triggered:
            self._handler.emit(record)
        else:
            self.buffered_records.append(record)
            if self._buffer_full:
                self.buffered_records.popleft()
            elif self.buffer_size and len(self.buffered_records) >= self.buffer_size:
                self._buffer_full = True
            return record.level >= self._level
        return False

    def rollover(self, record):
        if self._handler is None:
            self._handler = self._handler_factory(record, self)
        self._handler.emit_batch(iter(self.buffered_records), "escalation")
        self.buffered_records.clear()
        self._action_triggered = not self._reset

    @property
    def triggered(self):
        """This attribute is `True` when the action was triggered.  From
        this point onwards the finger crossed handler transparently
        forwards all log records to the inner handler.  If the handler resets
        itself this will always be `False`.
        """
        return self._action_triggered

    def emit(self, record):
        self.lock.acquire()
        try:
            if self.enqueue(record):
                self.rollover(record)
        finally:
            self.lock.release()


class GroupHandler(WrapperHandler):
    """A handler that buffers all messages until it is popped again and then
    forwards all messages to another handler.  This is useful if you for
    example have an application that does computations and only a result
    mail is required.  A group handler makes sure that only one mail is sent
    and not multiple.  Some other handles might support this as well, though
    currently none of the builtins do.

    Example::

        with GroupHandler(MailHandler(...)):
            # everything here ends up in the mail
            pass

    The :class:`GroupHandler` is implemented as a :class:`WrapperHandler`
    thus forwarding all attributes of the wrapper handler.

    Notice that this handler really only emit the records when the handler
    is popped from the stack.

    .. versionadded:: 0.3
    """

    _direct_attrs = frozenset(["handler", "pull_information", "buffered_records"])

    def __init__(self, handler, pull_information=True):
        WrapperHandler.__init__(self, handler)
        self.pull_information = pull_information
        self.buffered_records = []

    def rollover(self):
        self.handler.emit_batch(self.buffered_records, "group")
        self.buffered_records = []

    def pop_application(self):
        Handler.pop_application(self)
        self.rollover()

    @deprecated("Use pop_context instead")
    def pop_thread(self):
        Handler.pop_context(self)
        self.rollover()

    def pop_context(self):
        Handler.pop_context(self)
        self.rollover()

    @deprecated("Use pop_context instead")
    def pop_greenlet(self):
        Handler.pop_context(self)
        self.rollover()

    def emit(self, record):
        if self.pull_information:
            record.pull_information()
        self.buffered_records.append(record)



"""
logbook._termcolors
~~~~~~~~~~~~~~~~~~~

Provides terminal color mappings.

:copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
:license: BSD, see LICENSE for more details.
"""

esc = "\x1b["

codes = {"": "", "reset": esc + "39;49;00m"}

dark_colors = [
    "black",
    "darkred",
    "darkgreen",
    "brown",
    "darkblue",
    "purple",
    "teal",
    "lightgray",
]
light_colors = [
    "darkgray",
    "red",
    "green",
    "yellow",
    "blue",
    "fuchsia",
    "turquoise",
    "white",
]

x = 30
for d, l in zip(dark_colors, light_colors):  # noqa: E741
    codes[d] = esc + "%im" % x  # noqa: UP031
    codes[l] = esc + "%i;01m" % x  # noqa: UP031
    x += 1

del d, l, x

codes["darkteal"] = codes["turquoise"]
codes["darkyellow"] = codes["brown"]
codes["fuscia"] = codes["fuchsia"]


def _str_to_type(obj, strtype):
    """Helper for ansiformat and colorize"""
    if isinstance(obj, type(strtype)):
        return obj
    return obj.encode("ascii")


def colorize(color_key, text):
    """Returns an ANSI formatted text with the given color."""
    return (
        _str_to_type(codes[color_key], text) + text + _str_to_type(codes["reset"], text)
    )



import functools
import sys
import threading
from contextlib import contextmanager

from .base import DEBUG, Logger


class _SlowContextNotifier:
    def __init__(self, threshold, func):
        self.timer = threading.Timer(threshold, func)

    def __enter__(self):
        self.timer.start()
        return self

    def __exit__(self, *_):
        self.timer.cancel()


_slow_logger = Logger("Slow")


def logged_if_slow(*args, **kwargs):
    """Context manager that logs if operations within take longer than
    `threshold` seconds.

    :param threshold: Number of seconds (or fractions thereof) allwoed before
                      logging occurs. The default is 1 second.
    :param logger: :class:`~logbook.Logger` to use. The default is a 'slow'
                   logger.
    :param level: Log level. The default is `DEBUG`.
    :param func: (Deprecated). Function to call to perform logging.

    The remaining parameters are passed to the
    :meth:`~logbook.base.LoggerMixin.log` method.
    """
    threshold = kwargs.pop("threshold", 1)
    func = kwargs.pop("func", None)
    if func is None:
        logger = kwargs.pop("logger", _slow_logger)
        level = kwargs.pop("level", DEBUG)
        func = functools.partial(logger.log, level, *args, **kwargs)
    else:
        if "logger" in kwargs or "level" in kwargs:
            raise TypeError(
                "If using deprecated func parameter, 'logger' and"
                " 'level' arguments cannot be passed."
            )
        func = functools.partial(func, *args, **kwargs)

    return _SlowContextNotifier(threshold, func)


class _Local(threading.local):
    enabled = True


_local = _Local()


@contextmanager
def suppressed_deprecations():
    """Disables deprecation messages temporarily

    >>> with suppressed_deprecations():
    ...    call_some_deprecated_logic()

    .. versionadded:: 0.12
    """
    prev_enabled = _local.enabled
    _local.enabled = False
    try:
        yield
    finally:
        _local.enabled = prev_enabled


_deprecation_logger = Logger("deprecation")
_deprecation_locations = set()


def forget_deprecation_locations():
    _deprecation_locations.clear()


def _write_deprecations_if_needed(message, frame_correction):
    if not _local.enabled:
        return
    caller_location = _get_caller_location(frame_correction=frame_correction + 1)
    if caller_location not in _deprecation_locations:
        _deprecation_logger.warning(message, frame_correction=frame_correction + 1)
        _deprecation_locations.add(caller_location)


def log_deprecation_message(message, frame_correction=0):
    _write_deprecations_if_needed(
        f"Deprecation message: {message}", frame_correction=frame_correction + 1
    )


class _DeprecatedFunction:
    def __init__(self, func, message, obj=None, objtype=None):
        super().__init__()
        self._func = func
        self._message = message
        self._obj = obj
        self._objtype = objtype

    def _get_underlying_func(self):
        returned = self._func
        if isinstance(returned, classmethod):
            if hasattr(returned, "__func__"):
                returned = returned.__func__
            else:
                returned = returned.__get__(self._objtype).__func__
        return returned

    def __call__(self, *args, **kwargs):
        func = self._get_underlying_func()
        warning = f"{self._get_func_str()} is deprecated."
        if self._message is not None:
            warning += f" {self._message}"
        _write_deprecations_if_needed(warning, frame_correction=+1)
        if self._obj is not None:
            return func(self._obj, *args, **kwargs)
        elif self._objtype is not None:
            return func(self._objtype, *args, **kwargs)
        return func(*args, **kwargs)

    def _get_func_str(self):
        func = self._get_underlying_func()
        if self._objtype is not None:
            return f"{self._objtype.__name__}.{func.__name__}"
        return f"{func.__module__}.{func.__name__}"

    def __get__(self, obj, objtype):
        return self.bound_to(obj, objtype)

    def bound_to(self, obj, objtype):
        return _DeprecatedFunction(self._func, self._message, obj=obj, objtype=objtype)

    @property
    def __name__(self):
        return self._get_underlying_func().__name__

    @property
    def __doc__(self):
        if returned := self._get_underlying_func().__doc__:  # pylint: disable=no-member
            returned += "\n.. deprecated\n"  # pylint: disable=no-member
            if self._message:
                returned += f"   {self._message}"  # pylint: disable=no-member
        return returned

    @__doc__.setter
    def __doc__(self, doc):
        self._get_underlying_func().__doc__ = doc


def deprecated(func=None, message=None):
    """Marks the specified function as deprecated, and emits a warning when
    it's called.

    >>> @deprecated(message='No longer supported')
    ... def deprecated_func():
    ...     pass

    This will cause a warning log to be emitted when the function gets called,
    with the correct filename/lineno.

    .. versionadded:: 0.12
    """
    if isinstance(func, str):
        assert message is None
        message = func
        func = None

    if func is None:
        return functools.partial(deprecated, message=message)

    return _DeprecatedFunction(func, message)


def _get_caller_location(frame_correction):
    frame = sys._getframe(frame_correction + 1)  # pylint: disable=protected-access
    try:
        return (frame.f_code.co_name, frame.f_lineno)
    finally:
        del frame



"""
logbook.notifiers
~~~~~~~~~~~~~~~~~

System notify handlers for OSX and Linux.

:copyright: (c) 2010 by Armin Ronacher, Christopher Grebs.
:license: BSD, see LICENSE for more details.
"""

import base64
import os
import sys
from http import client as http_client
from time import time
from urllib.parse import urlencode

from typing_extensions import deprecated

from logbook.base import ERROR, NOTSET, WARNING
from logbook.handlers import Handler, LimitingHandlerMixin
from logbook.helpers import get_application_name


@deprecated("create_notification_handler is deprecated")
def create_notification_handler(application_name=None, level=NOTSET, icon=None):
    """Creates a handler perfectly fit the current platform.  On Linux
    systems this creates a :class:`LibNotifyHandler`, on OS X systems it
    will create a :class:`GrowlHandler`.

    .. deprecated:: 1.10
    """
    if sys.platform == "darwin":
        return GrowlHandler(application_name, level=level, icon=icon)
    return LibNotifyHandler(application_name, level=level, icon=icon)


class NotificationBaseHandler(Handler, LimitingHandlerMixin):
    """Baseclass for notification handlers."""

    def __init__(
        self,
        application_name=None,
        record_limit=None,
        record_delta=None,
        level=NOTSET,
        filter=None,
        bubble=False,
    ):
        Handler.__init__(self, level, filter, bubble)
        LimitingHandlerMixin.__init__(self, record_limit, record_delta)
        if application_name is None:
            application_name = get_application_name()
        self.application_name = application_name

    def make_title(self, record):
        """Called to get the title from the record."""
        return f"{record.channel}: {record.level_name.title()}"

    def make_text(self, record):
        """Called to get the text of the record."""
        return record.message


@deprecated("GrowlHandler is deprecated")
class GrowlHandler(NotificationBaseHandler):
    """A handler that dispatches to Growl.  Requires that either growl-py or
    py-Growl are installed.

    .. deprecated:: 1.9
    """

    def __init__(
        self,
        application_name=None,
        icon=None,
        host=None,
        password=None,
        record_limit=None,
        record_delta=None,
        level=NOTSET,
        filter=None,
        bubble=False,
    ):
        NotificationBaseHandler.__init__(
            self, application_name, record_limit, record_delta, level, filter, bubble
        )

        # growl is using the deprecated md5 module, but we really don't need
        # to see that deprecation warning
        from warnings import filterwarnings

        filterwarnings(module="Growl", category=DeprecationWarning, action="ignore")

        try:
            import Growl

            self._growl = Growl
        except ImportError:
            raise RuntimeError(
                "The growl module is not available.  You have "
                "to install either growl-py or py-Growl to "
                "use the GrowlHandler."
            )

        if icon is not None:
            if not os.path.isfile(icon):
                raise OSError("Filename to an icon expected.")
            icon = self._growl.Image.imageFromPath(icon)
        else:
            try:
                icon = self._growl.Image.imageWithIconForCurrentApplication()
            except TypeError:
                icon = None

        self._notifier = self._growl.GrowlNotifier(
            applicationName=self.application_name,
            applicationIcon=icon,
            notifications=[
                "Notset",
                "Debug",
                "Info",
                "Notice",
                "Warning",
                "Error",
                "Critical",
            ],
            hostname=host,
            password=password,
        )
        self._notifier.register()

    def is_sticky(self, record):
        """Returns `True` if the sticky flag should be set for this record.
        The default implementation marks errors and criticals sticky.
        """
        return record.level >= ERROR

    def get_priority(self, record):
        """Returns the priority flag for Growl.  Errors and criticals are
        get highest priority (2), warnings get higher priority (1) and the
        rest gets 0.  Growl allows values between -2 and 2.
        """
        if record.level >= ERROR:
            return 2
        elif record.level == WARNING:
            return 1
        return 0

    def emit(self, record):
        if not self.check_delivery(record)[1]:
            return
        self._notifier.notify(
            record.level_name.title(),
            self.make_title(record),
            self.make_text(record),
            sticky=self.is_sticky(record),
            priority=self.get_priority(record),
        )


@deprecated("LibNotifyHandler is deprecated")
class LibNotifyHandler(NotificationBaseHandler):
    """A handler that dispatches to libnotify.  Requires pynotify installed.
    If `no_init` is set to `True` the initialization of libnotify is skipped.

    .. deprecated:: 1.10
    """

    def __init__(
        self,
        application_name=None,
        icon=None,
        no_init=False,
        record_limit=None,
        record_delta=None,
        level=NOTSET,
        filter=None,
        bubble=False,
    ):
        NotificationBaseHandler.__init__(
            self, application_name, record_limit, record_delta, level, filter, bubble
        )

        try:
            import pynotify

            self._pynotify = pynotify
        except ImportError:
            raise RuntimeError(
                "The pynotify library is required for the LibNotifyHandler."
            )

        self.icon = icon
        if not no_init:
            pynotify.init(self.application_name)

    def set_notifier_icon(self, notifier, icon):
        """Used to attach an icon on a notifier object."""
        try:
            from gtk import gdk
        except ImportError:
            # TODO: raise a warning?
            raise RuntimeError("The gtk.gdk module is required to set an icon.")

        if icon is not None:
            if not isinstance(icon, gdk.Pixbuf):
                icon = gdk.pixbuf_new_from_file(icon)
            notifier.set_icon_from_pixbuf(icon)

    def get_expires(self, record):
        """Returns either EXPIRES_DEFAULT or EXPIRES_NEVER for this record.
        The default implementation marks errors and criticals as EXPIRES_NEVER.
        """
        pn = self._pynotify
        return pn.EXPIRES_NEVER if record.level >= ERROR else pn.EXPIRES_DEFAULT

    def get_urgency(self, record):
        """Returns the urgency flag for pynotify.  Errors and criticals are
        get highest urgency (CRITICAL), warnings get higher priority (NORMAL)
        and the rest gets LOW.
        """
        pn = self._pynotify
        if record.level >= ERROR:
            return pn.URGENCY_CRITICAL
        elif record.level == WARNING:
            return pn.URGENCY_NORMAL
        return pn.URGENCY_LOW

    def emit(self, record):
        if not self.check_delivery(record)[1]:
            return
        notifier = self._pynotify.Notification(
            self.make_title(record), self.make_text(record)
        )
        notifier.set_urgency(self.get_urgency(record))
        notifier.set_timeout(self.get_expires(record))
        self.set_notifier_icon(notifier, self.icon)
        notifier.show()


@deprecated("BoxcarHandler is deprecated")
class BoxcarHandler(NotificationBaseHandler):
    """Sends notifications to boxcar.io.  Can be forwarded to your iPhone or
    other compatible device.

    .. deprecated:: 1.9
    """

    api_url = "https://boxcar.io/notifications/"

    def __init__(
        self,
        email,
        password,
        record_limit=None,
        record_delta=None,
        level=NOTSET,
        filter=None,
        bubble=False,
    ):
        NotificationBaseHandler.__init__(
            self, None, record_limit, record_delta, level, filter, bubble
        )
        self.email = email
        self.password = password

    def get_screen_name(self, record):
        """Returns the value of the screen name field."""
        return record.level_name.title()

    def emit(self, record):
        if not self.check_delivery(record)[1]:
            return
        body = urlencode(
            {
                "notification[from_screen_name]": self.get_screen_name(record).encode(
                    "utf-8"
                ),
                "notification[message]": self.make_text(record).encode("utf-8"),
                "notification[from_remote_service_id]": str(int(time() * 100)),
            }
        )
        con = http_client.HTTPSConnection("boxcar.io")
        con.request(
            "POST",
            "/notifications/",
            headers={
                "Authorization": "Basic "
                + base64.b64encode(f"{self.email}:{self.password}".encode()).strip(),
            },
            body=body,
        )
        con.close()


@deprecated("NotifoHandler is deprecated")
class NotifoHandler(NotificationBaseHandler):
    """Sends notifications to notifo.com.  Can be forwarded to your Desktop,
    iPhone, or other compatible device.

    .. deprecated:: 1.9
    """

    def __init__(
        self,
        application_name=None,
        username=None,
        secret=None,
        record_limit=None,
        record_delta=None,
        level=NOTSET,
        filter=None,
        bubble=False,
        hide_level=False,
    ):
        try:
            import notifo
        except ImportError:
            raise RuntimeError(
                "The notifo module is not available.  You have "
                "to install notifo to use the NotifoHandler."
            )
        NotificationBaseHandler.__init__(
            self, None, record_limit, record_delta, level, filter, bubble
        )
        self._notifo = notifo
        self.application_name = application_name
        self.username = username
        self.secret = secret
        self.hide_level = hide_level

    def emit(self, record):
        if self.hide_level:
            _level_name = None
        else:
            _level_name = self.level_name

        self._notifo.send_notification(
            self.username,
            self.secret,
            None,
            record.message,
            self.application_name,
            _level_name,
            None,
        )


class PushoverHandler(NotificationBaseHandler):
    """Sends notifications to pushover.net.  Can be forwarded to your Desktop,
    iPhone, or other compatible device. If `priority` is not one of -2, -1, 0,
    or 1, it is set to 0 automatically.
    """

    def __init__(
        self,
        application_name=None,
        apikey=None,
        userkey=None,
        device=None,
        priority=0,
        sound=None,
        record_limit=None,
        record_delta=None,
        level=NOTSET,
        filter=None,
        bubble=False,
        max_title_len=100,
        max_message_len=512,
    ):
        super().__init__(None, record_limit, record_delta, level, filter, bubble)

        self.application_name = application_name
        self.apikey = apikey
        self.userkey = userkey
        self.device = device
        self.priority = priority
        self.sound = sound

        self.max_title_len = max_title_len
        self.max_message_len = max_message_len

        if self.application_name is None:
            self.title = None
        else:
            self.title = self._crop(self.application_name, self.max_title_len)

        if self.priority not in [-2, -1, 0, 1]:
            self.priority = 0

    def _crop(self, msg, max_len):
        if max_len is not None and max_len > 0 and len(msg) > max_len:
            return f"{msg[: max_len - 3]}..."
        else:
            return msg

    def emit(self, record):
        message = self._crop(record.message, self.max_message_len)

        body_dict = {
            "token": self.apikey,
            "user": self.userkey,
            "message": message,
            "priority": self.priority,
        }

        if self.title is not None:
            body_dict["title"] = self.title
        if self.device is not None:
            body_dict["device"] = self.device
        if self.sound is not None:
            body_dict["sound"] = self.sound

        body = urlencode(body_dict)
        con = http_client.HTTPSConnection("api.pushover.net")
        con.request("POST", "/1/messages.json", body=body)
        con.close()



"""
logbook.helpers
~~~~~~~~~~~~~~~

Various helper functions

:copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
:license: BSD, see LICENSE for more details.
"""

import errno
import os
import random
import re
import sys
import time
from collections.abc import Callable, Iterator
from datetime import datetime, timedelta, timezone
from typing import TypeVar

# this regexp also matches incompatible dates like 20070101 because
# some libraries (like the python xmlrpclib modules) use this
_iso8601_re = re.compile(
    # date
    r"(\d{4})(?:-?(\d{2})(?:-?(\d{2}))?)?"
    # time
    r"(?:T(\d{2}):(\d{2})(?::(\d{2}(?:\.\d+)?))?(Z|[+-]\d{2}:\d{2})?)?$"
)
_missing = object()

T = TypeVar("T")


can_rename_open_file = False
if os.name == "nt":
    try:
        import ctypes

        _MOVEFILE_REPLACE_EXISTING = 0x1
        _MOVEFILE_WRITE_THROUGH = 0x8
        _MoveFileEx = ctypes.windll.kernel32.MoveFileExW

        def _rename(src, dst):
            if _rename_atomic(src, dst):
                return True
            retry = 0
            rv = False
            while not rv and retry < 100:
                rv = _MoveFileEx(
                    src, dst, _MOVEFILE_REPLACE_EXISTING | _MOVEFILE_WRITE_THROUGH
                )
                if not rv:
                    time.sleep(0.001)
                    retry += 1
            return rv

        # new in Vista and Windows Server 2008
        _CreateTransaction = ctypes.windll.ktmw32.CreateTransaction
        _CommitTransaction = ctypes.windll.ktmw32.CommitTransaction
        _MoveFileTransacted = ctypes.windll.kernel32.MoveFileTransactedW
        _CloseHandle = ctypes.windll.kernel32.CloseHandle
        can_rename_open_file = True

        def _rename_atomic(src, dst):
            ta = _CreateTransaction(None, 0, 0, 0, 0, 1000, "Logbook rename")
            if ta == -1:
                return False
            try:
                retry = 0
                rv = False
                while not rv and retry < 100:
                    if rv := _MoveFileTransacted(
                        src,
                        dst,
                        None,
                        None,
                        _MOVEFILE_REPLACE_EXISTING | _MOVEFILE_WRITE_THROUGH,
                        ta,
                    ):
                        rv = _CommitTransaction(ta)
                        break
                    else:
                        time.sleep(0.001)
                        retry += 1
                return rv
            finally:
                _CloseHandle(ta)

    except Exception:

        def _rename(src, dst):
            return False

        def _rename_atomic(src, dst):
            return False

    def rename(src, dst):
        # Try atomic or pseudo-atomic rename
        if _rename(src, dst):
            return
        # Fall back to "move away and replace"
        try:
            os.rename(src, dst)
        except OSError:
            e = sys.exc_info()[1]
            if e.errno not in (errno.EEXIST, errno.EACCES):
                raise
            old = f"{dst}-{random.randint(0, 2**31 - 1):08x}"
            os.rename(dst, old)
            os.rename(src, dst)
            try:
                os.unlink(old)
            except Exception:
                pass

else:
    rename = os.rename
    can_rename_open_file = True

_JSON_SIMPLE_TYPES = (bool, float, int, str)


def to_safe_json(data):
    """Makes a data structure safe for JSON silently discarding invalid
    objects from nested structures.  This also converts dates.
    """

    def _convert(obj):
        if obj is None:
            return None
        elif isinstance(obj, _JSON_SIMPLE_TYPES):
            return obj
        elif isinstance(obj, datetime):
            return format_iso8601(obj)
        elif isinstance(obj, list):
            return [_convert(x) for x in obj]
        elif isinstance(obj, tuple):
            return tuple(_convert(x) for x in obj)
        elif isinstance(obj, dict):
            rv = {}
            for key, value in obj.items():
                if not isinstance(key, str):
                    key = str(key)
                rv[key] = _convert(value)
            return rv

    return _convert(data)


if sys.version_info >= (3, 12):

    def datetime_utcnow():
        """datetime.utcnow() but doesn't emit a deprecation warning.

        Will be fixed by https://github.com/getlogbook/logbook/issues/353
        """
        return datetime.now(timezone.utc).replace(tzinfo=None)

    def datetime_utcfromtimestamp(timestamp):
        """datetime.utcfromtimesetamp() but doesn't emit a deprecation warning.

        Will be fixed by https://github.com/getlogbook/logbook/issues/353
        """
        return datetime.fromtimestamp(timestamp, timezone.utc).replace(tzinfo=None)

else:
    datetime_utcnow = datetime.utcnow
    datetime_utcfromtimestamp = datetime.utcfromtimestamp


def format_iso8601(d=None):
    """Returns a date in iso8601 format."""
    if d is None:
        d = datetime_utcnow()
    rv = d.strftime("%Y-%m-%dT%H:%M:%S")
    if d.microsecond:
        rv += "." + str(d.microsecond)
    return rv + "Z"


def parse_iso8601(value):
    """Parse an iso8601 date into a datetime object.  The timezone is
    normalized to UTC.
    """
    m = _iso8601_re.match(value)
    if m is None:
        raise ValueError("not a valid iso8601 date value")

    groups = m.groups()
    args = []
    for group in groups[:-2]:
        if group is not None:
            group = int(group)
        args.append(group)
    seconds = groups[-2]
    if seconds is not None:
        if "." in seconds:
            sec, usec = seconds.split(".")
            args.append(int(sec))
            args.append(int(usec.ljust(6, "0")))
        else:
            args.append(int(seconds))

    rv = datetime(*args)
    tz = groups[-1]
    if tz and tz != "Z":
        args = [int(x) for x in tz[1:].split(":")]
        delta = timedelta(hours=args[0], minutes=args[1])
        if tz[0] == "+":
            rv -= delta
        else:
            rv += delta

    return rv


def get_application_name():
    if not sys.argv or not sys.argv[0]:
        return "Python"
    return os.path.basename(sys.argv[0]).title()


class cached_property:
    """A property that is lazily calculated and then cached."""

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def get_iterator_next_method(it: Iterator[T]) -> Callable[[], T]:
    return lambda: next(it)



"""
logbook.compat
~~~~~~~~~~~~~~

Backwards compatibility with stdlib's logging package and the
warnings module.

:copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
:license: BSD, see LICENSE for more details.
"""

import logging
import sys
import warnings
from collections.abc import Mapping
from datetime import timezone

import logbook

from .helpers import datetime_utcfromtimestamp


def redirect_logging(set_root_logger_level=True):
    """Permanently redirects logging to the stdlib.  This also
    removes all otherwise registered handlers on root logger of
    the logging system but leaves the other loggers untouched.

    :param set_root_logger_level: controls of the default level of the legacy
       root logger is changed so that all legacy log messages get redirected
       to Logbook
    """
    del logging.root.handlers[:]
    logging.root.addHandler(RedirectLoggingHandler())
    if set_root_logger_level:
        logging.root.setLevel(logging.DEBUG)


class redirected_logging:
    """Temporarily redirects logging for all threads and reverts
    it later to the old handlers.  Mainly used by the internal
    unittests::

        from logbook.compat import redirected_logging

        with redirected_logging():
            ...
    """

    def __init__(self, set_root_logger_level=True):
        self.old_handlers = logging.root.handlers[:]
        self.old_level = logging.root.level
        self.set_root_logger_level = set_root_logger_level

    def start(self):
        redirect_logging(self.set_root_logger_level)

    def end(self, etype=None, evalue=None, tb=None):
        logging.root.handlers[:] = self.old_handlers
        logging.root.setLevel(self.old_level)

    __enter__ = start
    __exit__ = end


class LoggingCompatRecord(logbook.LogRecord):
    def _format_message(self, msg, *args, **kwargs):
        if kwargs:
            assert not args
            return msg % kwargs
        else:
            assert not kwargs
            return msg % tuple(args)


class RedirectLoggingHandler(logging.Handler):
    """A handler for the stdlib's logging system that redirects
    transparently to logbook.  This is used by the
    :func:`redirect_logging` and :func:`redirected_logging`
    functions.

    If you want to customize the redirecting you can subclass it.
    """

    def __init__(self):
        logging.Handler.__init__(self)

    def convert_level(self, level):
        """Converts a logging level into a logbook level."""
        if level >= logging.CRITICAL:
            return logbook.CRITICAL
        if level >= logging.ERROR:
            return logbook.ERROR
        if level >= logging.WARNING:
            return logbook.WARNING
        if level >= logging.INFO:
            return logbook.INFO
        return logbook.DEBUG

    def find_extra(self, old_record):
        """Tries to find custom data from the old logging record.  The
        return value is a dictionary that is merged with the log record
        extra dictionaries.
        """
        rv = vars(old_record).copy()
        for key in (
            "name",
            "msg",
            "args",
            "levelname",
            "levelno",
            "pathname",
            "filename",
            "module",
            "exc_info",
            "exc_text",
            "lineno",
            "funcName",
            "created",
            "msecs",
            "relativeCreated",
            "thread",
            "threadName",
            "greenlet",
            "processName",
            "process",
        ):
            rv.pop(key, None)
        return rv

    def find_caller(self, old_record):
        """Tries to find the caller that issued the call."""
        frm = sys._getframe(2)
        while frm is not None:
            if (
                frm.f_globals is globals()
                or frm.f_globals is logbook.base.__dict__
                or frm.f_globals is logging.__dict__
            ):
                frm = frm.f_back
            else:
                return frm

    def convert_time(self, timestamp):
        """Converts the UNIX timestamp of the old record into a
        datetime object as used by logbook.
        """
        return datetime_utcfromtimestamp(timestamp)

    def convert_record(self, old_record):
        """Converts an old logging record into a logbook log record."""
        args = old_record.args
        kwargs = None

        # Logging allows passing a mapping object, in which case args will be a mapping.
        if isinstance(args, Mapping):
            kwargs = args
            args = None
        record = LoggingCompatRecord(
            old_record.name,
            self.convert_level(old_record.levelno),
            old_record.msg,
            args,
            kwargs,
            old_record.exc_info,
            self.find_extra(old_record),
            self.find_caller(old_record),
        )
        record.time = self.convert_time(old_record.created)
        return record

    def emit(self, record):
        logbook.dispatch_record(self.convert_record(record))


class LoggingHandler(logbook.Handler):
    """Does the opposite of the :class:`RedirectLoggingHandler`, it sends
    messages from logbook to logging.  Because of that, it's a very bad
    idea to configure both.

    This handler is for logbook and will pass stuff over to a logger
    from the standard library.

    Example usage::

        from logbook.compat import LoggingHandler, warn

        with LoggingHandler():
            warn("This goes to logging")
    """

    def __init__(self, logger=None, level=logbook.NOTSET, filter=None, bubble=False):
        logbook.Handler.__init__(self, level, filter, bubble)
        if logger is None:
            logger = logging.getLogger()
        elif isinstance(logger, str):
            logger = logging.getLogger(logger)
        self.logger = logger

    def get_logger(self, record):
        """Returns the logger to use for this record.  This implementation
        always return :attr:`logger`.
        """
        return self.logger

    def convert_level(self, level):
        """Converts a logbook level into a logging level."""
        if level >= logbook.CRITICAL:
            return logging.CRITICAL
        if level >= logbook.ERROR:
            return logging.ERROR
        if level >= logbook.WARNING:
            return logging.WARNING
        if level >= logbook.INFO:
            return logging.INFO
        return logging.DEBUG

    def convert_time(self, dt):
        """Converts a datetime object into a timestamp."""
        if dt.tzinfo is None:
            # Logbook uses naive datetimes to represent UTC (utcnow)
            return dt.replace(tzinfo=timezone.utc).timestamp()
        return dt.timestamp()

    def convert_record(self, old_record):
        """Converts a record from logbook to logging."""
        record = logging.LogRecord(
            old_record.channel,
            self.convert_level(old_record.level),
            old_record.filename,
            old_record.lineno,
            old_record.message,
            (),
            old_record.exc_info,
            func=old_record.func_name,
        )
        for key, value in old_record.extra.items():
            record.__dict__.setdefault(key, value)
        record.created = self.convert_time(old_record.time)
        return record

    def emit(self, record):
        self.get_logger(record).handle(self.convert_record(record))


def redirect_warnings():
    """Like :func:`redirected_warnings` but will redirect all warnings
    to the shutdown of the interpreter:

    .. code-block:: python

        from logbook.compat import redirect_warnings

        redirect_warnings()
    """
    redirected_warnings().__enter__()


class redirected_warnings:
    """A context manager that copies and restores the warnings filter upon
    exiting the context, and logs warnings using the logbook system.

    The :attr:`~logbook.LogRecord.channel` attribute of the log record will be
    the import name of the warning.

    Example usage:

    .. code-block:: python

        from logbook.compat import redirected_warnings
        from warnings import warn

        with redirected_warnings():
            warn(DeprecationWarning("logging should be deprecated"))
    """

    def __init__(self):
        self._entered = False

    def message_to_unicode(self, message):
        return str(message)

    def make_record(self, message, exception, filename, lineno):
        category = exception.__name__
        if exception.__module__ not in ("exceptions", "builtins"):
            category = exception.__module__ + "." + category
        rv = logbook.LogRecord(category, logbook.WARNING, message)
        # we don't know the caller, but we get that information from the
        # warning system.  Just attach them.
        rv.filename = filename
        rv.lineno = lineno
        return rv

    def start(self):
        if self._entered:  # pragma: no cover
            raise RuntimeError("Cannot enter %r twice" % self)  # noqa: UP031
        self._entered = True
        self._filters = warnings.filters
        warnings.filters = self._filters[:]
        self._showwarning = warnings.showwarning

        def showwarning(message, category, filename, lineno, file=None, line=None):
            message = self.message_to_unicode(message)
            record = self.make_record(message, category, filename, lineno)
            logbook.dispatch_record(record)

        warnings.showwarning = showwarning

    def end(self, etype=None, evalue=None, tb=None):
        if not self._entered:  # pragma: no cover
            raise RuntimeError("Cannot exit %r without entering first" % self)  # noqa: UP031
        warnings.filters = self._filters
        warnings.showwarning = self._showwarning

    __enter__ = start
    __exit__ = end



"""
logbook
~~~~~~~

Simple logging library that aims to support desktop, command line
and web applications alike.

:copyright: (c) 2010 by Armin Ronacher, Georg Brandl.
:license: BSD, see LICENSE for more details.
"""

import os
from typing import Any

from .base import (
    CRITICAL,
    DEBUG,
    ERROR,
    INFO,
    NOTICE,
    NOTSET,
    TRACE,
    WARNING,
    Flags,
    Logger,
    LoggerGroup,
    LogRecord,
    NestedSetup,
    Processor,
    dispatch_record,
    get_level_name,
    lookup_level,
    set_datetime_format,
)
from .handlers import (
    BrotliCompressionHandler,
    FileHandler,
    FingersCrossedHandler,
    GMailHandler,
    GroupHandler,
    GZIPCompressionHandler,
    Handler,
    HashingHandlerMixin,
    LimitingHandlerMixin,
    MailHandler,
    MonitoringFileHandler,
    NTEventLogHandler,
    NullHandler,
    RotatingFileHandler,
    StderrHandler,
    StreamHandler,
    StringFormatter,
    StringFormatterHandlerMixin,
    SyslogHandler,
    TestHandler,
    TimedRotatingFileHandler,
    WrapperHandler,
    create_syshandler,
)

# create an anonymous default logger and provide all important
# methods of that logger as global functions
_default_logger = Logger("Generic")
_default_logger.suppress_dispatcher = True
trace = _default_logger.trace
debug = _default_logger.debug
info = _default_logger.info
warn = _default_logger.warn
warning = _default_logger.warning
notice = _default_logger.notice
error = _default_logger.error
exception = _default_logger.exception
catch_exceptions = _default_logger.catch_exceptions
critical = _default_logger.critical
log = _default_logger.log
del _default_logger


# install a default global handler
if os.environ.get("LOGBOOK_INSTALL_DEFAULT_HANDLER"):
    default_handler = StderrHandler()
    default_handler.push_application()

__all__ = (
    "CRITICAL",
    "DEBUG",
    "ERROR",
    "INFO",
    "NOTICE",
    "NOTSET",
    "TRACE",
    "WARNING",
    "BrotliCompressionHandler",
    "FileHandler",
    "FingersCrossedHandler",
    "Flags",
    "GMailHandler",
    "GZIPCompressionHandler",
    "GroupHandler",
    "Handler",
    "HashingHandlerMixin",
    "LimitingHandlerMixin",
    "LogRecord",
    "Logger",
    "LoggerGroup",
    "MailHandler",
    "MonitoringFileHandler",
    "NTEventLogHandler",
    "NestedSetup",
    "NullHandler",
    "Processor",
    "RotatingFileHandler",
    "StderrHandler",
    "StreamHandler",
    "StringFormatter",
    "StringFormatterHandlerMixin",
    "SyslogHandler",
    "TestHandler",
    "TimedRotatingFileHandler",
    "WrapperHandler",
    "catch_exceptions",
    "create_syshandler",
    "critical",
    "debug",
    "dispatch_record",
    "error",
    "exception",
    "get_level_name",
    "info",
    "log",
    "lookup_level",
    "notice",
    "set_datetime_format",
    "trace",
    "warn",
    "warning",
)


def __getattr__(name: str) -> Any:
    if name != "__version__":
        msg = f"module '{__name__}' has no attribute '{name}'"
        raise AttributeError(msg)

    import warnings
    from importlib.metadata import version

    warnings.warn(
        "logbook.__version__ is deprecated and will be removed in a "
        "future release. Use importlib.metadata instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    return version("Logbook")
